# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import django.db.models.deletion
import touchtechnology.common.db.models
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("competition", "0006_sportingpulse_import_fields"),
    ]

    operations = [
        migrations.CreateModel(
            name="MatchVideo",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("url", models.URLField(max_length=255)),
                (
                    "approved_by",
                    touchtechnology.common.db.models.ForeignKey(
                        blank=True,
                        to=settings.AUTH_USER_MODEL,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                    ),
                ),
                (
                    "match",
                    touchtechnology.common.db.models.ForeignKey(
                        related_name="videos",
                        to="competition.Match",
                        on_delete=django.db.models.deletion.PROTECT,
                    ),
                ),
            ],
            options={},
            bases=(models.Model,),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2018-07-04 22:02
from __future__ import unicode_literals

import django.db.models.deletion
import touchtechnology.common.db.models
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('competition', '0029_seasonreferee'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='seasonreferee',
            options={'ordering': ('club', 'season', 'person__last_name', 'person__first_name'), 'verbose_name': 'referee'},
        ),
        migrations.AddField(
            model_name='match',
            name='referees',
            field=touchtechnology.common.db.models.ManyToManyField(blank=True, related_name='matches', to='competition.SeasonReferee'),
        ),
    ]



from django.template import Context, Template
from example_app.tests.factories import RelativeFactory
from test_plus import TestCase


class RelatedFilterTest(TestCase):
    def setUp(self):
        RelativeFactory.reset_sequence()
        self.template = Template(
            """
        {% load common mvp_tags %}
        {% spaceless %}
            <dl>
                {% for manager, name in object|related:related %}
                    <dt>{{ name }}</dt>
                    <dd>{{ manager|type }} ({{ manager.count }})</dd>
                {% endfor %}
            </dl>
        {% endspaceless %}
        """
        )

    def _get_context(self, obj, rel=None):
        if rel is None:
            rel = ()
        context = {"object": obj, "related": rel}
        return Context(context)

    def test_related_no_whitelist(self):
        rel = RelativeFactory()
        context = self._get_context(rel.link)
        output = self.template.render(context).strip()
        self.assertEqual(output, "<dl></dl>")

    def test_related_empty_whitelist(self):
        rel = RelativeFactory()
        context = self._get_context(rel.link, ())
        output = self.template.render(context).strip()
        self.assertEqual(output, "<dl></dl>")

    def test_related_specific_whitelist(self):
        rel = RelativeFactory()
        context = self._get_context(rel.link, ("relative_set",),)
        output = self.template.render(context).strip()
        self.assertEqual(
            output, "<dl>" "<dt>relative_set</dt>" "<dd>RelatedManager (1)</dd>" "</dl>"
        )



import logging
import socket
import sys

import django
import pytz
from django.utils import timezone
from django.utils.deprecation import MiddlewareMixin
from touchtechnology.common.utils import get_timezone_from_request

logger = logging.getLogger(__name__)


class AcceptsMiddleware(MiddlewareMixin):
    def process_request(self, request):
        request.accepts = set(request.META.get("HTTP_ACCEPT", "").split(","))


def served_by_middleware(get_response):
    """
    Expose in our HTTP response headers a number of useful environment values
    that may be useful during debugging.
    """
    def middleware(request):
        response = get_response(request)
        if 'HTTP_X_DETAIL' in request.META or response.status_code >= 500:
            response['X-Django-Version'] = django.get_version()
            response['X-Python-Version'] = ' '.join(sys.version.split())
            response['X-Pytz-Version'] = pytz.__version__
            response['X-Served-By'] = socket.gethostname()
            response['X-Timezone-Name'] = timezone.get_current_timezone_name()
        return response
    return middleware


def timezone_middleware(get_response):
    """
    Activate the timezone for this request/response cycle if
    get_timezone_from_request returns a tzinfo.
    """
    def middleware(request):
        tzinfo = get_timezone_from_request(request)

        if tzinfo is not None:
            logger.debug(
                'Activating user timezone "%s" for this request.', tzinfo)
            request.tzinfo = tzinfo
            timezone.activate(tzinfo)

        response = get_response(request)

        if tzinfo is not None:
            logger.debug(
                'Forcibly reset the timezone from "%s" back to "%s".',
                request.tzinfo, timezone.get_default_timezone_name())
            timezone.deactivate()

        return response
    return middleware



# -*- coding: utf-8 -*-
# Generated by Django 1.11.5 on 2017-09-25 21:29
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('content', '0003_auto_20160831_0608'),
        ('common', '0005_sitemapnode_populate_kwargs'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='placeholderkeywordargument',
            unique_together=set([]),
        ),
        migrations.RemoveField(
            model_name='placeholderkeywordargument',
            name='node',
        ),
        migrations.DeleteModel(
            name='PlaceholderKeywordArgument',
        ),
    ]



from django.core.paginator import Paginator
from django.template import Library
from django.utils.translation import gettext_lazy as _

from touchtechnology.news.models import Article, Category

register = Library()


@register.filter("category")
def get_category(slug):
    return Category.objects.get(slug=slug)


@register.inclusion_tag("touchtechnology/news/_related_list.html", takes_context=True)
def related_articles(context, article, limit=None, order_by=None):
    categories = article.categories.live()
    articles = (
        Article.objects.live()
        .exclude(pk=article.pk)
        .filter(categories__in=categories)
        .distinct()
    )

    if order_by is not None:
        articles = articles.order_by(*order_by.split(","))

    if limit is not None:
        articles = articles[: int(limit)]

    # FIXME backwards compatibility for custom templates
    context["slice"] = ":"

    context["article_list"] = articles
    return context


@register.inclusion_tag("touchtechnology/news/_related_list.html", takes_context=True)
def related_categories(context, article=None, limit=None):
    """
    If an article is provided, then we select categories relating to it.
    Otherwise we select all article categories.
    """
    if article is None:
        categories = Category.objects.all()
    else:
        categories = article.categories.all()
    context["category_list"] = categories
    return context


@register.inclusion_tag(
    "touchtechnology/news/_latest_articles.html", takes_context=True
)
def latest_articles(context, count=5, title=_("Latest News")):
    articles = Article.objects.live()
    paginator = Paginator(articles, count)
    page = paginator.page(1)
    context["paginator"] = paginator
    context["page"] = page
    context["article_list"] = page.object_list
    context["title"] = title
    return context



# Generated by Django 3.2.13 on 2022-07-25 12:25

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("competition", "0039_season_live_stream"),
    ]

    operations = [
        migrations.AddField(
            model_name="season",
            name="live_stream_privacy",
            field=models.CharField(
                choices=[
                    ("public", "Public"),
                    ("private", "Private"),
                    ("unlisted", "Unlisted"),
                ],
                default="private",
                max_length=20,
            ),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import django.core.validators
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('competition', '0012_hashtag_include_symbol'),
    ]

    operations = [
        migrations.AlterField(
            model_name='simplescorematchstatistic',
            name='played',
            field=models.SmallIntegerField(default=0, blank=True, validators=[django.core.validators.MinValueValidator(0), django.core.validators.MaxValueValidator(1)]),
        ),
    ]



from django.db import migrations

import touchtechnology.common.db.models


class Migration(migrations.Migration):

    dependencies = [
        ("news", "0005_auto_20191122_1340"),
    ]

    operations = [
        migrations.AddField(
            model_name="article",
            name="copy",
            field=touchtechnology.common.db.models.HTMLField(
                blank=True, verbose_name="Copy"
            ),
        ),
    ]



import django.db.models.deletion
from django.db import migrations, models

import touchtechnology.admin.mixins
import touchtechnology.common.db.models


class Migration(migrations.Migration):

    dependencies = [
        ("news", "0008_delete_articlecontent"),
    ]

    operations = [
        migrations.CreateModel(
            name="Translation",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("locale", models.CharField(max_length=10)),
                ("headline", models.CharField(max_length=150, verbose_name="Headline")),
                ("abstract", models.TextField(verbose_name="Abstract")),
                (
                    "copy",
                    touchtechnology.common.db.models.HTMLField(
                        blank=True, verbose_name="Copy"
                    ),
                ),
                (
                    "article",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="translations",
                        to="news.Article",
                    ),
                ),
            ],
            options={"unique_together": {("article", "locale")},},
            bases=(touchtechnology.admin.mixins.AdminUrlMixin, models.Model),
        ),
    ]



from __future__ import unicode_literals

import functools
import logging
import operator
from datetime import timedelta
from operator import or_

import httplib2
import pytz
from apiclient.discovery import build
from dateutil.relativedelta import relativedelta
from django import forms
from django.conf import settings
from django.contrib import messages
from django.contrib.sitemaps import views as sitemaps_views
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Case, Count, F, Q, Sum, When
from django.http import Http404, HttpResponse, HttpResponseGone
from django.shortcuts import get_object_or_404
from django.urls import include, path, re_path, reverse
from django.urls.exceptions import NoReverseMatch
from django.utils import timezone
from django.utils.module_loading import import_string
from django.utils.safestring import mark_safe
from django.utils.translation import gettext, gettext_lazy as _
from googleapiclient.errors import HttpError
from guardian.utils import get_40x_or_None
from icalendar import Calendar, Event
from oauth2client.contrib.django_util import get_storage

from touchtechnology.common.decorators import login_required_m
from touchtechnology.common.sites import Application
from touchtechnology.common.utils import get_perms_for_model
from tournamentcontrol.competition.dashboard import (
    matches_require_basic_results,
    matches_require_details_results,
)
from tournamentcontrol.competition.decorators import competition_by_slug_m
from tournamentcontrol.competition.forms import (
    ConfigurationForm,
    DivisionTournamentScheduleForm,
    MatchResultFormSet,
    MatchStatisticFormset,
    MultiConfigurationForm,
    RankingConfigurationForm,
)
from tournamentcontrol.competition.models import (
    Competition,
    Match,
    Person,
    SimpleScoreMatchStatistic,
)
from tournamentcontrol.competition.rank import (
    DayView as RankDay,
    DivisionView as RankDivision,
    IndexView as RankIndex,
    MonthView as RankMonth,
    TeamView as RankTeam,
    YearView as RankYear,
)
from tournamentcontrol.competition.utils import (
    FauxQueryset,
    team_needs_progressing,
)

LOG = logging.getLogger(__name__)


def permissions_required(
    request,
    model,
    instance=None,
    return_403=True,
    accept_global_perms=True,
    create=False,
    perms=None,
):
    # If no perms are specified, build sensible default using built in
    # permission types that come batteries included with Django.
    if perms is None:
        perms = get_perms_for_model(model, change=True)

        # When we're doing a creation we should have permission to create the object.
        if create:
            perms = get_perms_for_model(model, add=True)

    # Determine the user's permission to edit this object using the
    # get_40x_or_None - saves decorating view method with
    # @permission_required_or_403
    has_permission = get_40x_or_None(
        request,
        perms,
        obj=instance,
        return_403=return_403,
        accept_global_perms=accept_global_perms,
    )

    if has_permission is not None:
        return has_permission


class CompetitionAdminMixin(object):
    """
    Some administrative functions should be allowed to authenticated users on
    the front-end of the site. As long as we only use primitives from the base
    ``touchtechnology.common.sites.Application`` class it will integrate fine
    with both.
    """

    def day_runsheet(
        self, request, season, date, extra_context, templates=None, **kwargs
    ):
        matches = season.matches.filter(date=date).order_by(
            "is_bye", "datetime", "play_at"
        )
        templates = self.template_path(
            "runsheet.html", season.slug, season.competition.slug
        )
        return self.generic_list(
            request,
            matches,
            templates=templates,
            paginate_by=0,
            permission_required=False,
            extra_context=extra_context,
        )

    def match_results(
        self,
        request,
        competition,
        season,
        date,
        extra_context,
        redirect_to,
        division=None,
        time=None,
        **kwargs,
    ):
        matches = Match.objects.filter(
            stage__division__season=season,
            stage__division__season__competition=competition,
            date=date,
        )

        if division is not None:
            matches = matches.filter(stage__division=division)

        # ensure only matches with progressed teams are able to be updated
        matches = matches.exclude(team_needs_progressing, is_bye=False)
        matches = matches.exclude(home_team__isnull=True, away_team__isnull=True)

        # FIXME too complex, be verbose so we can all read and understand it
        if time is not None:
            base = Q(time=time)
            bye_kwargs = matches.filter(time=time).values("stage", "round")
            time_or_byes = functools.reduce(
                operator.or_, [Q(time__isnull=True, **kw) for kw in bye_kwargs], base
            )
            matches = matches.filter(time_or_byes)
        else:
            matches = matches.order_by("date", "time", "play_at")

        match_queryset = matches.filter(is_bye=False)
        bye_queryset = matches.filter(is_bye=True)

        if request.method == "POST":
            match_formset = MatchResultFormSet(
                data=request.POST, queryset=match_queryset, prefix="matches"
            )
            bye_formset = MatchResultFormSet(
                data=request.POST, queryset=bye_queryset, prefix="byes"
            )

            if match_formset.is_valid() and bye_formset.is_valid():
                match_formset.save()
                bye_formset.save()
                messages.success(request, _("Your changes have been saved."))
                return self.redirect(redirect_to)
        else:
            match_formset = MatchResultFormSet(
                queryset=match_queryset, prefix="matches"
            )
            bye_formset = MatchResultFormSet(queryset=bye_queryset, prefix="byes")

        context = {
            "competition": competition,
            "season": season,
            "date": date,
            "division": division,
            "match_formset": match_formset,
            "bye_formset": bye_formset,
            "matches": matches,
            "cancel_url": redirect_to,
        }
        context.update(extra_context)

        templates = self.template_path("match_results.html")
        return self.render(request, templates, context)

    def stream_control(
        self,
        request,
        competition,
        season,
        date,
        extra_context,
        redirect_to,
        time,
        **kwargs,
    ):
        # FIXME: absolute hack view in order to get this going in time for Euros
        match_queryset = Match.objects.filter(
            stage__division__season=season,
            stage__division__season__competition=competition,
            date=date,
            time=time,
            external_identifier__isnull=False,
        ).order_by("play_at")

        class StreamControlForm(forms.Form):
            status = forms.ChoiceField(
                label="Live Streaming Control",
                choices=(
                    ("live", "Start streaming"),
                    ("complete", "Stop streaming"),
                ),
                widget=forms.RadioSelect,
                required=True,
                help_text=mark_safe(
                    "Start streams about 1 minute before matches start.<br/>"
                    "Stop streams about 1 minute after matches conclude."
                ),
            )

            def save(slf):
                storage = get_storage(request)
                credentials = storage.get()
                if credentials is None or credentials.invalid:
                    return self.redirect(reverse("google_oauth:authorize"))

                youtube = build(
                    "youtube", "v3", http=credentials.authorize(httplib2.Http())
                )

                broadcast_status = slf.cleaned_data["status"]

                for match in match_queryset:
                    try:
                        youtube.liveBroadcasts().transition(
                            broadcastStatus=broadcast_status,
                            id=match.external_identifier,
                            part="snippet,status",
                        ).execute()
                    except HttpError as exc:
                        messages.error(request, f"{match}: {exc.reason}")
                    else:
                        messages.success(
                            request, f"{match}: broadcast is {broadcast_status!r}"
                        )

        if request.method == "POST":
            form = StreamControlForm(data=request.POST)

            if form.is_valid():
                res = form.save()
                if isinstance(res, HttpResponse):
                    return res
                return self.redirect(redirect_to)
        else:
            form = StreamControlForm()

        context = {
            "competition": competition,
            "season": season,
            "date": date,
            "form": form,
            "matches": match_queryset,
            "cancel_url": redirect_to,
        }
        context.update(extra_context)

        templates = self.template_path("stream_control.html")
        return self.render(request, templates, context)

    def edit_match_detail(
        self, request, stage, match, extra_context, redirect_to, **kwargs
    ):
        conditions = {
            "home_team__isnull": False,
            "away_team__isnull": False,
            "home_team_score__isnull": False,
            "away_team_score__isnull": False,
        }

        if match is None:
            match = get_object_or_404(stage.matches, pk=match.pk, **conditions)

        def team_faux_queryset(team):
            stats = FauxQueryset(SimpleScoreMatchStatistic, team=team)
            for player in team.people.filter(is_player=True):
                try:
                    statistic = SimpleScoreMatchStatistic.objects.get(
                        match=match, player=player.person
                    )
                except ObjectDoesNotExist:
                    statistic = SimpleScoreMatchStatistic(
                        match=match,
                        player=player.person,
                        number=player.number,
                        played=1,
                    )
                stats.append(statistic)
            return stats

        home_queryset = team_faux_queryset(match.home_team)
        away_queryset = team_faux_queryset(match.away_team)

        if request.method == "POST":
            home = MatchStatisticFormset(
                data=request.POST,
                score=match.home_team_score,
                prefix="home",
                queryset=home_queryset,
            )

            away = MatchStatisticFormset(
                data=request.POST,
                score=match.away_team_score,
                prefix="away",
                queryset=away_queryset,
            )

            if home.is_valid() and away.is_valid():
                home.save()
                away.save()
                messages.success(request, _("Your changes have been saved."))
                return self.redirect(redirect_to)

        else:
            home = MatchStatisticFormset(
                prefix="home", score=match.home_team_score, queryset=home_queryset
            )
            away = MatchStatisticFormset(
                prefix="away", score=match.away_team_score, queryset=away_queryset
            )

        context = {
            "object": match,
            "formsets": (home, away),
        }
        context.update(extra_context)

        templates = self.template_path("match_detail.html")
        return self.render(request, templates, context)


class CompetitionSite(CompetitionAdminMixin, Application):
    kwargs_form_class = ConfigurationForm

    def __init__(self, name="competition", app_name="competition", **kwargs):
        # store the node for future reference
        self.node = kwargs.get("node")
        super().__init__(name=name, app_name=app_name, **kwargs)
        self._competitions = Competition.objects.filter(enabled=True)
        if "competition" in kwargs and kwargs["competition"]:
            self._competitions = self._competitions.filter(slug=kwargs["competition"])

    def result_urls(self):
        return [
            path("", self.results, name="results"),
            path("match/<int:match>/", self.edit_match_detail, name="match-details"),
            re_path(r"^(?P<datestr>\d{8})/$", self.results, name="results"),
            re_path(
                r"^(?P<datestr>\d{8})/(?P<timestr>\d{4})/$",
                self.match_results,
                name="match-results",
            ),
        ]

    def runsheet_urls(self):
        return [
            path("", self.runsheet, name="runsheet"),
            re_path(r"^(?P<datestr>\d{8})/$", self.day_runsheet, name="runsheet"),
        ]

    def stream_urls(self):
        return [
            path("", self.stream, name="stream"),
            re_path(r"^(?P<datestr>\d{8})/$", self.stream, name="stream"),
            re_path(
                r"^(?P<datestr>\d{8})/(?P<timestr>\d{4})/$",
                self.stream_control,
                name="stream-control",
            ),
        ]

    def season_urls(self):
        return [
            path("", self.season, name="season"),
            path("forfeit/", self.forfeit_list, name="forfeit-list"),
            path("forfeit/<int:match>/", self.forfeit, name="forfeit"),
            path("videos/", self.season_videos, name="season-videos"),
            path("club:<slug:club>/", self.club, name="club"),
            path("club:<slug:club>.ics", self.calendar, name="calendar"),
            path("results/", include(self.result_urls())),
            path("runsheet/", include(self.runsheet_urls())),
            path("stream/", include(self.stream_urls())),
            path("<slug:division>.ics", self.calendar, name="calendar"),
            path("<slug:division>/", self.division, name="division"),
            path("<slug:division>:<slug:stage>/", self.stage, name="stage"),
            path("<slug:division>:<slug:stage>:<slug:pool>/", self.pool, name="pool"),
            path("<slug:division>/match:<int:match>/", self.match, name="match"),
            path(
                "<slug:division>/match:<int:match>/video/",
                self.match_video,
                name="match-video",
            ),
            path("<slug:division>/<slug:team>.ics", self.calendar, name="calendar"),
            path("<slug:division>/<slug:team>/", self.team, name="team"),
        ]

    def competition_urls(self):
        return [
            path("", self.competition, name="competition"),
            path("<slug:season>.ics", self.calendar, name="calendar"),
            path("<slug:season>/", include(self.season_urls())),
        ]

    def get_urls(self):
        if self.kwargs.get("season") and self.kwargs.get("competition"):
            urlpatterns = [
                path("", include(self.season_urls()), kwargs=self.kwargs),
            ]
        elif self.kwargs.get("competition"):
            urlpatterns = [
                path("", include(self.competition_urls()), kwargs=self.kwargs),
            ]
        else:
            urlpatterns = [
                path("", self.index, name="index"),
                path("<slug:competition>/", include(self.competition_urls())),
            ]
        return urlpatterns

    def sitemap_index(
        self,
        request,
        node=None,
        competition=None,
        season=None,
        division=None,
        *args,
        **kwargs,
    ):
        # We need to "pop" the node keyword argument
        return sitemaps_views.index(request, *args, **kwargs)

    def sitemap_section(
        self,
        request,
        node=None,
        competition=None,
        season=None,
        division=None,
        *args,
        **kwargs,
    ):
        # We need to "pop" the node keyword argument
        return sitemaps_views.sitemap(request, *args, **kwargs)

    @property
    def urls(self):
        return self.get_urls(), self.app_name, self.name

    @property
    def competitions(self):
        return self._competitions.prefetch_related("seasons")

    def index(self, request, **kwargs):
        return self.generic_list(
            request,
            self._competitions,
            templates=self.template_path("index.html"),
            paginate_by=self._competitions.count(),
            extra_context=kwargs,
        )

    @competition_by_slug_m
    def competition(self, request, competition, extra_context, **kwargs):
        templates = self.template_path("competition.html", competition.slug)
        return self.generic_detail(
            request,
            self._competitions,
            slug=competition.slug,
            templates=templates,
            extra_context=extra_context,
        )

    @competition_by_slug_m
    def season(self, request, competition, season, extra_context, **kwargs):
        templates = self.template_path("season.html", competition.slug, season.slug)
        extra_context.update(
            season.matches.exclude(is_bye=True).aggregate(
                timeslot_count=Count("datetime", distinct=True),
                match_count=Count("pk", distinct=True),
                points_scored=Sum(
                    Case(
                        When(
                            statistics__match__stage__division__season=season,
                            then=F("statistics__points"),
                        )
                    )
                ),
                caps_awarded=Sum(
                    Case(
                        When(
                            statistics__match__stage__division__season=season,
                            then=F("statistics__played"),
                        )
                    )
                ),
            )
        )
        return self.generic_detail(
            request,
            competition.seasons,
            slug=season.slug,
            templates=templates,
            extra_context=extra_context,
        )

    @competition_by_slug_m
    def runsheet(self, request, competition, season, extra_context, **kwargs):
        context = {"dates": season.matches.dates("date", "day")}
        context.update(extra_context)
        templates = self.template_path(
            "runsheet_list.html", competition.slug, season.slug
        )
        return self.render(request, templates, context)

    @competition_by_slug_m
    def day_runsheet(self, request, season, date, extra_context, **kwargs):
        return super().day_runsheet(request, season, date, extra_context, **kwargs)

    @competition_by_slug_m
    @login_required_m
    def results(self, request, competition, season, extra_context, date=None, **kwargs):
        has_permission = permissions_required(request, Match, return_403=False)
        if has_permission is not None:
            return has_permission

        if date is not None:
            matches = season.matches.filter(date=date)
        else:
            matches = season.matches

        # Allow super-user accounts visibility to look 1 year into the future.
        if request.user.is_superuser:
            now = timezone.now() + relativedelta(years=1)
        else:
            now = None

        # Filter the list of matches to those which require result entry
        basic_results = matches_require_basic_results(now, matches)
        details_results = matches_require_details_results(matches, True)

        context = {
            "datetimes": [
                timezone.localtime(each, pytz.timezone(season.timezone))
                for each in basic_results.datetimes("datetime", "minute")
            ],
            "details": details_results,
        }
        context.update(extra_context)
        templates = self.template_path("results.html", competition.slug, season.slug)
        return self.render(request, templates, context)

    @competition_by_slug_m
    @login_required_m
    def stream(self, request, competition, season, extra_context, date=None, **kwargs):
        has_permission = permissions_required(request, Match, return_403=False)
        if has_permission is not None:
            return has_permission

        if date is not None:
            matches = season.matches.filter(date=date)
        else:
            matches = season.matches

        # Allow super-user accounts visibility to look 1 year into the future.
        if request.user.is_superuser:
            now = timezone.now() + relativedelta(years=1)
        else:
            now = None

        # Filter the list of matches to those which require streaming control
        stream_start = matches.exclude(external_identifier__isnull=True)

        context = {
            "datetimes": [
                timezone.localtime(each, pytz.timezone(season.timezone))
                for each in stream_start.datetimes("datetime", "minute")
            ],
        }
        context.update(extra_context)
        templates = self.template_path("stream.html", competition.slug, season.slug)
        return self.render(request, templates, context)

    @competition_by_slug_m
    @login_required_m
    def match_results(
        self, request, competition, season, date, time, extra_context, **kwargs
    ):
        has_permission = permissions_required(request, Match, return_403=False)
        if has_permission is not None:
            return has_permission
        redirect_to = self.reverse("results", args=(competition.slug, season.slug))
        return super().match_results(
            request,
            competition=competition,
            season=season,
            date=date,
            time=time,
            extra_context=extra_context,
            redirect_to=redirect_to,
            **kwargs,
        )

    @competition_by_slug_m
    @login_required_m
    def stream_control(
        self, request, competition, season, date, time, extra_context, **kwargs
    ):
        has_permission = permissions_required(request, Match, return_403=False)
        if has_permission is not None:
            return has_permission
        redirect_to = self.reverse("stream", args=(competition.slug, season.slug))
        return super().stream_control(
            request,
            competition=competition,
            season=season,
            date=date,
            time=time,
            extra_context=extra_context,
            redirect_to=redirect_to,
            **kwargs,
        )

    @login_required_m
    @competition_by_slug_m
    def edit_match_detail(self, request, match, extra_context, **kwargs):
        has_permission = permissions_required(
            request, Match, instance=match, return_403=False
        )
        if has_permission is not None:
            return has_permission
        redirect_to = self.reverse(
            "results",
            args=(
                match.stage.division.season.competition.slug,
                match.stage.division.season.slug,
            ),
        )
        return super().edit_match_detail(
            request,
            stage=match.stage,
            match=match,
            extra_context=extra_context,
            redirect_to=redirect_to,
            **kwargs,
        )

    @competition_by_slug_m
    def season_videos(self, request, competition, season, extra_context, **kwargs):
        templates = self.template_path(
            "season_videos.html", competition.slug, season.slug
        )
        queryset = (
            Match.objects.select_related(
                "stage__division",
                "home_team__club",
                "away_team__club",
                "play_at",
            )
            .exclude(videos__isnull=True)
            .filter(stage__division__season=season)
            .order_by("datetime", "play_at")
        )
        return self.generic_list(
            request,
            queryset,
            paginate_by=1000,
            templates=templates,
            extra_context=extra_context,
        )

    @competition_by_slug_m
    def club(self, request, competition, season, club, extra_context, **kwargs):
        templates = self.template_path(
            "club.html", competition.slug, season.slug, club.slug
        )
        return self.generic_detail(
            request,
            season.clubs,
            slug=club.slug,
            templates=templates,
            extra_context=extra_context,
        )

    @competition_by_slug_m
    def division(self, request, competition, season, division, extra_context, **kwargs):
        templates = self.template_path(
            "division.html", competition.slug, season.slug, division.slug
        )
        extra_context.update(
            division.matches.exclude(is_bye=True).aggregate(
                timeslot_count=Count("datetime", distinct=True),
                match_count=Count("pk", distinct=True),
                points_scored=Sum(
                    Case(
                        When(
                            statistics__match__stage__division=division,
                            then=F("statistics__points"),
                        )
                    )
                ),
                caps_awarded=Sum(
                    Case(
                        When(
                            statistics__match__stage__division=division,
                            then=F("statistics__played"),
                        )
                    )
                ),
            )
        )
        return self.generic_detail(
            request,
            season.divisions,
            slug=division.slug,
            templates=templates,
            extra_context=extra_context,
        )

    @competition_by_slug_m
    def stage(
        self, request, competition, season, division, stage, extra_context, **kwargs
    ):
        templates = self.template_path(
            "stage.html", competition.slug, season.slug, division.slug, stage.slug
        )
        extra_context["parent"] = stage
        extra_context.update(
            stage.matches.exclude(is_bye=True).aggregate(
                timeslot_count=Count("datetime", distinct=True),
                match_count=Count("pk", distinct=True),
                points_scored=Sum(
                    Case(
                        When(
                            statistics__match__stage=stage, then=F("statistics__points")
                        )
                    )
                ),
                caps_awarded=Sum(
                    Case(
                        When(
                            statistics__match__stage=stage, then=F("statistics__played")
                        )
                    )
                ),
            )
        )
        return self.generic_detail(
            request,
            division.stages,
            slug=stage.slug,
            templates=templates,
            extra_context=extra_context,
        )

    @competition_by_slug_m
    def pool(
        self,
        request,
        competition,
        season,
        division,
        stage,
        pool,
        extra_context,
        **kwargs,
    ):
        # FIXME sadly the template name was not changed when we refactored
        #       pools to be subordinates of Stage.
        templates = self.template_path(
            "divisiongroup.html",
            competition.slug,
            season.slug,
            division.slug,
            stage.slug,
            pool.slug,
        )
        extra_context.update(
            pool.matches.exclude(is_bye=True).aggregate(
                timeslot_count=Count("datetime", distinct=True),
                match_count=Count("pk", distinct=True),
                points_scored=Sum(
                    Case(
                        When(
                            statistics__match__stage_group=pool,
                            then=F("statistics__points"),
                        )
                    )
                ),
                caps_awarded=Sum(
                    Case(
                        When(
                            statistics__match__stage_group=pool,
                            then=F("statistics__played"),
                        )
                    )
                ),
            )
        )
        return self.generic_detail(
            request,
            stage.pools,
            slug=pool.slug,
            templates=templates,
            extra_context=extra_context,
        )

    @competition_by_slug_m
    def team(
        self, request, competition, season, division, team, extra_context, **kwargs
    ):
        templates = self.template_path(
            "team.html", competition.slug, season.slug, division.slug, team.slug
        )

        extra_context.update(
            team.matches.exclude(is_bye=True).aggregate(
                timeslot_count=Count("datetime", distinct=True),
                match_count=Count("pk", distinct=True),
                points_scored=Sum(
                    Case(
                        When(
                            statistics__player__teamassociation__team=team,
                            then=F("statistics__points"),
                        )
                    )
                ),
                caps_awarded=Sum(
                    Case(
                        When(
                            statistics__player__teamassociation__team=team,
                            then=F("statistics__played"),
                        )
                    )
                ),
            )
        )
        return self.generic_detail(
            request,
            division.teams,
            slug=team.slug,
            templates=templates,
            extra_context=extra_context,
        )

    @competition_by_slug_m
    def calendar(self, request, season, club=None, division=None, team=None, **kwargs):
        if season.disable_calendar:
            # The GONE response informs client that they should remove this resource
            # from their cache. When a calendar has been added to user's mobile device
            # they may never look at it again, but we continue to process the requests
            # which can have poor performance. Try to influence a cleanup of clients.
            return HttpResponseGone()

        if team is not None:
            matches = team.matches
        elif division is not None:
            matches = division.matches
        elif club is not None:
            matches = club.matches.filter(stage__division__season=season)
        else:
            matches = season.matches

        # Do not include matches which have not had the time scheduled
        matches = matches.exclude(datetime__isnull=True)

        # Perform select_related to reduce extra queries
        matches = matches.select_related("stage__division__season__competition")

        # Reduce the size of the data set to return from the database
        matches = matches.defer(
            "date",
            "time",
            "home_team__copy",
            "home_team__names_locked",
            "home_team__order",
            "home_team__short_title",
            "home_team__timeslots_after",
            "home_team__timeslots_before",
            "home_team_score",
            "away_team__copy",
            "away_team__names_locked",
            "away_team__order",
            "away_team__short_title",
            "away_team__timeslots_after",
            "away_team__timeslots_before",
            "away_team_score",
            "bye_processed",
            "evaluated",
            "external_identifier",
            "forfeit_winner",
            "include_in_ladder",
            "is_bye",
            "is_forfeit",
            "is_washout",
            "rank_importance",
            "videos",
            "stage__division__season__competition__copy",
            "stage__division__season__competition__enabled",
            "stage__division__season__competition__order",
            "stage__division__season__competition__rank_importance",
            "stage__division__season__competition__short_title",
            "stage__division__season__competition__short_title",
            "stage__division__season__competition__slug_locked",
            "stage__division__season__competition__slug_locked",
            "stage__division__season__competition__title",
            "stage__division__season__complete",
            "stage__division__season__copy",
            "stage__division__season__disable_calendar",
            "stage__division__season__enabled",
            "stage__division__season__hashtag",
            "stage__division__season__mode",
            "stage__division__season__mvp_results_public",
            "stage__division__season__order",
            "stage__division__season__rank_importance",
            "stage__division__season__short_title",
            "stage__division__season__slug_locked",
            "stage__division__season__start_date",
            "stage__division__season__statistics",
            "stage__division__season__timezone",
            "stage__division__bonus_points_formula",
            "stage__division__copy",
            "stage__division__draft",
            "stage__division__forfeit_against_score",
            "stage__division__forfeit_for_score",
            "stage__division__games_per_day",
            "stage__division__include_forfeits_in_played",
            "stage__division__order",
            "stage__division__points_formula",
            "stage__division__rank_division",
            "stage__division__rank_importance",
            "stage__division__short_title",
            "stage__division__slug_locked",
            "stage__division__sportingpulse_url",
            "stage__carry_ladder",
            "stage__copy",
            "stage__follows",
            "stage__keep_ladder",
            "stage__keep_mvp",
            "stage__order",
            "stage__rank_importance",
            "stage__scale_group_points",
            "stage__short_title",
            "stage__slug_locked",
        )

        # Remove any matches that are part of a draft division unless being viewed
        # by a superuser.
        if not request.user.is_superuser:
            matches = matches.exclude(stage__division__draft=True)

        # For development server turn back plain text to make debugging easier
        if settings.DEBUG:
            content_type = "text/plain"
        else:
            content_type = "text/calendar"

        response = HttpResponse(content_type=content_type)

        cal = Calendar()
        cal.add("prodid", "-//Tournament Control//%s//" % request.get_host())
        cal.add("version", "2.0")

        for match in matches.order_by("datetime", "play_at"):
            event = Event()
            event["uid"] = match.uuid.hex
            event.add("summary", match.title)
            event.add("location", f"{match.stage.division.title} ({match.stage.title})")
            event.add("dtstart", match.datetime)

            # FIXME match duration should not be hardcoded
            event.add("dtend", match.datetime + timedelta(minutes=45))
            # FIXME should be the last modified time of the match
            event.add("dtstamp", timezone.now())

            try:
                # Determine the resource uri to the detailed match view
                uri = reverse(
                    "competition:match",
                    kwargs={
                        "match": match.pk,
                        "division": match.stage.division.slug,
                        "season": match.stage.division.season.slug,
                        "competition": match.stage.division.season.competition.slug,
                    },
                )
            except NoReverseMatch:
                LOG.exception("Unable to resolve url for %r", match)
            else:
                # Combine the resource uri with our current request context
                event.add("description", request.build_absolute_uri(uri))

            cal.add_component(event)

        response.write(cal.to_ical())
        return response

    @competition_by_slug_m
    def match(
        self, request, competition, season, division, match, extra_context, **kwargs
    ):
        if match.is_bye or match.home_team is None or match.away_team is None:
            # Temporary, realistically nobody should ever have this URL but we
            # will use GONE so that Google removes it from their index.
            return HttpResponseGone()
        templates = self.template_path(
            "match.html", competition.slug, season.slug, division.slug
        )
        return self.generic_detail(
            request,
            division.matches,
            pk=match.pk,
            templates=templates,
            extra_context=extra_context,
        )

    @competition_by_slug_m
    def match_video(
        self, request, competition, season, division, match, extra_context, **kwargs
    ):
        if match.is_bye or match.home_team is None or match.away_team is None:
            # Temporary, realistically nobody should ever have this URL but we
            # will use GONE so that Google removes it from their index.
            return HttpResponseGone()
        templates = self.template_path(
            "videos.html", competition.slug, season.slug, division.slug
        )
        return self.generic_detail(
            request,
            division.matches,
            pk=match.pk,
            templates=templates,
            extra_context=extra_context,
        )

        # TODO add related model to record videos
        # return self.generic_list(request, match.videos,
        #                          templates=templates,
        #                          extra_context=extra_context)

    @login_required_m
    @competition_by_slug_m
    def forfeit_list(self, request, competition, season, extra_context, **kwargs):
        """
        List the matches that the visitor is permitted to forfeit. Must only
        show matches for teams that they are a registered member of (either as
        a player or other role).
        """
        try:
            teams = request.user.person.teams
        except Person.DoesNotExist:
            raise Http404

        matches = season.matches.filter(Q(home_team=teams) | Q(away_team=teams))

        context = {
            "matches": matches,
        }
        context.update(extra_context)

        templates = self.template_path("forfeit_list.html")
        return self.render(request, templates, context)

    @login_required_m
    @competition_by_slug_m
    def forfeit(self, request, competition, season, extra_context, match, **kwargs):
        """
        View that will allow a team member to forfeit a match they are due to
        be playing in.
        """
        try:
            teams = request.user.person.teams
        except Person.DoesNotExist:
            raise Http404

        matches = season.matches.filter(Q(home_team=teams) | Q(away_team=teams))

        redirect = self.redirect(
            self.reverse(
                "forfeit-list",
                kwargs={
                    "competition": competition.slug,
                    "season": season.slug,
                },
            )
        )

        if match not in matches:
            messages.add_message(
                request,
                messages.ERROR,
                "Attempting to forfeit a match you are not involved in.",
            )
            return redirect

        # decide which team to forfeit
        if match.home_team in teams and match.away_team in teams:
            messages.add_message(
                request,
                messages.WARNING,
                _(
                    "You are a member of both teams for this match, please "
                    "contact the competition manager."
                ),
            )
            return redirect

        elif match.home_team in teams:
            team = match.home_team
        elif match.away_team in teams:
            team = match.away_team
        else:
            messages.add_message(
                request,
                messages.ERROR,
                _(
                    "You are not a member of either team, please contact the "
                    "competition manager."
                ),
            )
            return redirect

        if request.method == "POST":
            # use the forfeit API to advise who made lodged the forfeit
            UNABLE_TO_POST = _(
                "Unable to post forfeit, please contact the " "competition manager."
            )
            try:
                success = match.forfeit(team, request.user.person)
            except AssertionError:
                # logger.exception("Failed precondition checks")
                messages.add_message(request, messages.ERROR, UNABLE_TO_POST)

            if success:
                messages.add_message(
                    request,
                    messages.INFO,
                    _(
                        "The match has been forfeit, notifications will be sent "
                        "to affected parties shortly."
                    ),
                )
            else:
                messages.add_message(request, messages.ERROR, UNABLE_TO_POST)

            return redirect

        context = {
            "match": match,
            "team": team,
        }
        context.update(extra_context)

        templates = self.template_path("forfeit.html")
        return self.render(request, templates, context)


class MultiCompetitionSite(CompetitionSite):
    kwargs_form_class = MultiConfigurationForm

    @classmethod
    def verbose_name(cls):
        return gettext("Multiple Competitions")

    def __init__(self, name="competition", app_name="competition", **kwargs):
        self.node = kwargs.get("node")  # store the node for future reference
        super().__init__(name=name, app_name=app_name, **kwargs)
        self._competitions = Competition.objects.filter(enabled=True)
        if "competition" in kwargs:
            q = functools.reduce(or_, [Q(slug=slug) for slug in kwargs["competition"]])
            self._competitions = self._competitions.filter(q)

    def get_urls(self):
        urlpatterns = [
            path("", self.index, name="index"),
            path("<slug:competition>/", include(self.competition_urls())),
        ]
        return urlpatterns


class RegistrationSite(Application):
    def __init__(self, name="rego", app_name="rego", **kwargs):
        super().__init__(name=name, app_name=app_name, **kwargs)

    def get_urls(self):
        return []


class TournamentCalculatorSite(Application):
    def __init__(self, name="calculator", app_name="calculator", **kwargs):
        self.form_class = import_string(
            kwargs.pop(
                "form_class",
                "tournamentcontrol.competition.forms.TournamentScheduleForm",
            )
        )
        super().__init__(name=name, app_name=app_name, **kwargs)

    def get_urls(self):
        urlpatterns = [
            path("", self.index, kwargs={"form_class": self.form_class}, name="index"),
            path(
                "live/",
                self.index,
                kwargs={
                    "form_class": DivisionTournamentScheduleForm,
                    "template_name": "live.html",
                },
                name="division",
            ),
        ]
        return urlpatterns

    def index(self, request, form_class, template_name="index.html", **extra_context):
        context = {}

        if request.GET:
            form = form_class(data=request.GET)
        else:
            form = form_class()

        if form.is_valid():
            context.setdefault("data", form.get_context())

        context.setdefault("form", form)
        context.update(extra_context)

        templates = self.template_path(template_name)
        return self.render(request, templates, context)


class RankingSite(Application):
    kwargs_form_class = RankingConfigurationForm

    def __init__(self, name="ranking", app_name="ranking", **kwargs):
        super().__init__(name=name, app_name=app_name, **kwargs)

    def get_urls(self):
        urlpatterns = [
            path("", RankIndex.as_view(), name="index"),
            path("<int:year>/", RankYear.as_view(), name="year"),
            path("<int:year>/<str:month>/", RankMonth.as_view(), name="month"),
            path("<int:year>/<str:month>/<int:day>/", RankDay.as_view(), name="day"),
            path(
                "<int:year>/<str:month>/<int:day>/<slug:slug>/",
                RankDivision.as_view(),
                name="rank",
            ),
            path(
                "<int:year>/<str:month>/<int:day>/<slug:slug>/<slug:team>/",
                RankTeam.as_view(),
                name="team",
            ),
        ]
        return urlpatterns


competition = CompetitionSite()
registration = RegistrationSite()
calculator = TournamentCalculatorSite()
ranking = RankingSite()



from django.conf import settings

__all__ = (
    "DETAIL_IMAGE_KWARGS",
    "DETAIL_IMAGE_PROCESSORS",
    "PAGINATE_BY",
    "THUMBNAIL_IMAGE_KWARGS",
    "THUMBNAIL_IMAGE_PROCESSORS",
)

S = lambda n, d=None: getattr(settings, "TOUCHTECHNOLOGY_NEWS_" + n, d) or d

DEFAULT_DETAIL_IMAGE_KWARGS = {}
DEFAULT_DETAIL_IMAGE_PROCESSORS = (
    ("pilkit.processors.resize.SmartResize", (320, 240), {}),
)

DEFAULT_THUMBNAIL_IMAGE_KWARGS = {}
DEFAULT_THUMBNAIL_IMAGE_PROCESSORS = (
    ("pilkit.processors.resize.SmartResize", (160, 120), {}),
)

DETAIL_IMAGE_KWARGS = S("DETAIL_IMAGE_KWARGS", DEFAULT_DETAIL_IMAGE_KWARGS)
DETAIL_IMAGE_PROCESSORS = S("DETAIL_IMAGE_PROCESSORS", DEFAULT_DETAIL_IMAGE_PROCESSORS)
PAGINATE_BY = S("PAGINATE_BY", 5)
THUMBNAIL_IMAGE_KWARGS = S("THUMBNAIL_IMAGE_KWARGS", DEFAULT_THUMBNAIL_IMAGE_KWARGS)
THUMBNAIL_IMAGE_PROCESSORS = S(
    "THUMBNAIL_IMAGE_PROCESSORS", DEFAULT_THUMBNAIL_IMAGE_PROCESSORS
)



# Generated by Django 3.2.13 on 2022-07-24 01:52

from django.db import migrations
import touchtechnology.common.db.models


class Migration(migrations.Migration):

    dependencies = [
        ("competition", "0037_alter_match_evaluated"),
    ]

    operations = [
        migrations.AddField(
            model_name="match",
            name="live_stream",
            field=touchtechnology.common.db.models.BooleanField(default=False),
        ),
    ]



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import collections
import logging
import uuid
from datetime import datetime, timedelta
from decimal import Decimal, InvalidOperation

import django
import pytz
from cloudinary.models import CloudinaryField
from dateutil.relativedelta import relativedelta
from dateutil.rrule import MINUTELY, WEEKLY, rrule, rruleset
from django import forms
from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.postgres import fields as PG
from django.core import validators
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Count, DateField, DateTimeField, Q, Sum, TimeField
from django.db.models.deletion import CASCADE, PROTECT, SET_NULL
from django.template import Template
from django.template.loader import get_template
from django.utils import timezone
from django.utils.functional import cached_property, lazy
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from touchtechnology.admin.mixins import AdminUrlMixin as BaseAdminUrlMixin
from touchtechnology.common.db.models import (
    BooleanField,
    ForeignKey,
    HTMLField,
    LocationField,
    ManyToManyField,
)
from touchtechnology.common.models import SitemapNodeBase
from tournamentcontrol.competition.constants import (
    GENDER_CHOICES,
    LiveStreamPrivacy,
    PYTZ_TIME_ZONE_CHOICES,
    SEASON_MODE_CHOICES,
    WIN_LOSE,
)
from tournamentcontrol.competition.managers import (
    LadderEntryManager,
    MatchManager,
)
from tournamentcontrol.competition.mixins import ModelDiffMixin
from tournamentcontrol.competition.query import (
    DivisionQuerySet,
    StageQuerySet,
    StatisticQuerySet,
)
from tournamentcontrol.competition.signals import match_forfeit
from tournamentcontrol.competition.utils import (
    FauxQueryset,
    combine_and_localize,
    stage_group_position,
    stage_group_position_re,
    team_and_division,
)
from tournamentcontrol.competition.validators import validate_hashtag

logger = logging.getLogger(__name__)
lazy_get_template = lazy(get_template, Template)

match_title_tpl = lazy_get_template("tournamentcontrol/competition/_match_title.txt")

stage_group_position_tpl = lazy_get_template(
    "tournamentcontrol/competition/_stage_group_position.txt"
)

win_lose_team_tpl = lazy_get_template(
    "tournamentcontrol/competition/_win_lose_team.txt"
)


class AdminUrlMixin(BaseAdminUrlMixin):
    def _get_url_args(self):
        return (self.pk,)


class RankDivisionMixin(models.Model):
    rank_division_parent_attr = None

    rank_division = models.ForeignKey(
        "RankDivision", null=True, blank=True, on_delete=PROTECT
    )

    @property
    def rank_division_parent(self):
        return getattr(self, self.rank_division_parent_attr)

    def get_rank_division(self):
        if self.rank_division is not None:
            return self.rank_division
        if self.rank_division_parent_attr:
            return self.rank_division_parent.get_rank_division()

    class Meta:
        abstract = True


class RankImportanceMixin(models.Model):
    rank_importance_parent_attr = None

    rank_importance = models.DecimalField(
        max_digits=6, decimal_places=3, null=True, blank=True
    )

    @property
    def rank_importance_parent(self):
        if isinstance(self.rank_importance_parent_attr, str):
            return getattr(self, self.rank_importance_parent_attr)
        if isinstance(self.rank_importance_parent_attr, (list, tuple)):
            for attr in self.rank_importance_parent_attr:
                obj = getattr(self, attr)
                if obj is not None:
                    return obj

    def get_rank_importance(self):
        if self.rank_importance is not None:
            return self.rank_importance
        if self.rank_importance_parent_attr:
            return self.rank_importance_parent.get_rank_importance()

    class Meta:
        abstract = True


class TimeZoneChoiceField(forms.ChoiceField):
    """
    Simple customisation of a ``forms.ChoiceField`` that allows a default
    declaration of ``PYTZ_TIME_ZONE_CHOICES`` to fill the ``choices`` without
    the need to add this on the model field declaration.

    ``PYTZ_TIME_ZONE_CHOICES`` is dependant upon the currently installed
    version of ``pytz`` and this was affecting the migrations framework.
    """

    def __init__(self, max_length=None, *args, **kwargs):
        choices = kwargs.pop("choices", PYTZ_TIME_ZONE_CHOICES)
        if django.VERSION[:2] >= (1, 11):
            kwargs.pop("empty_value", None)
        super(TimeZoneChoiceField, self).__init__(choices=choices, *args, **kwargs)


class TimeZoneField(models.CharField):
    """
    Simple customisation of a ``models.CharField`` that allows a default
    declaration of ``PYTZ_TIME_ZONE_CHOICES`` to fill the ``choices`` without
    the need to add this on the model field declaration.

    ``PYTZ_TIME_ZONE_CHOICES`` is dependant upon the currently installed
    version of ``pytz`` and this was affecting the migrations framework.
    """

    def deconstruct(self):
        name, path, args, kwargs = super(TimeZoneField, self).deconstruct()
        return name, "django.db.models.CharField", args, kwargs

    def formfield(self, form_class=TimeZoneChoiceField, *args, **kwargs):
        required = kwargs.pop("required", settings.USE_TZ)
        return super(TimeZoneField, self).formfield(
            form_class=form_class, required=required, *args, **kwargs
        )


class LadderPointsField(models.TextField):
    def formfield(self, form_class=None, **kwargs):
        from tournamentcontrol.competition.forms import (
            LadderPointsField as LadderPointsFormField,
        )

        if form_class is None:
            form_class = LadderPointsFormField
        return super(LadderPointsField, self).formfield(form_class=form_class, **kwargs)


class TwitterField(models.CharField):
    def formfield(self, form_class=None, **kwargs):
        if form_class is None:
            form_class = forms.RegexField
            kwargs.setdefault("regex", r"^@[\w_]+$")
        return super(TwitterField, self).formfield(form_class=form_class, **kwargs)


class OrderedSitemapNode(SitemapNodeBase):
    copy = HTMLField(blank=True)
    order = models.PositiveIntegerField(default=1)

    def __repr__(self):
        return "<{}: {}>".format(self.__class__.__name__, self.title)

    def __str__(self):
        return self.title

    class Meta:
        abstract = True
        ordering = ("order",)


class Competition(AdminUrlMixin, RankImportanceMixin, OrderedSitemapNode):
    enabled = BooleanField(default=True)
    clubs = ManyToManyField("Club", blank=True, related_name="competitions")

    def _get_admin_namespace(self):
        return "admin:fixja:competition"

    class Meta(OrderedSitemapNode.Meta):
        pass


class Club(AdminUrlMixin, SitemapNodeBase):
    email = models.EmailField(max_length=255, blank=True)
    website = models.URLField(max_length=255, blank=True)
    twitter = TwitterField(
        max_length=50,
        blank=True,
        help_text=_("Official Twitter name for use in " 'social "mentions"'),
    )
    facebook = models.URLField(max_length=255, blank=True)
    youtube = models.URLField(max_length=255, blank=True)
    primary = ForeignKey(
        "Person",
        blank=True,
        null=True,
        related_name="+",
        verbose_name=_("Primary contact"),
        label_from_instance="get_full_name",
        help_text=_("Appears on the front-end with other " "club information."),
        on_delete=PROTECT,
    )
    primary_position = models.CharField(
        max_length=200,
        blank=True,
        verbose_name=_("Position"),
        help_text=_("Position of the primary " "contact"),
    )
    abbreviation = models.CharField(
        max_length=3,
        blank=True,
        help_text=_("Optional 3-letter " "abbreviation to be used on " "scoreboards."),
    )

    class Meta:
        ordering = ("title",)

    @cached_property
    def _mvp_related(self):
        """
        Supercedes _mvp_annotate due to GROUP BY exceptions. Take full control
        of the query. See the history for the earlier implementation.
        """
        query = """
            SELECT
                "competition_person"."uuid",
                "competition_person"."first_name",
                "competition_person"."last_name",
                "competition_person"."gender",
                "competition_person"."date_of_birth",
                "competition_person"."club_id",
                "competition_person"."user_id",
                SUM("competition_simplescorematchstatistic"."played") AS "stats_played",
                MIN("competition_match"."date") AS "debut",
                SUM("competition_simplescorematchstatistic"."points") AS "stats_points",
                COUNT("competition_teamassociation"."id") AS "teams_count",
                "%(user)s"."id"
            FROM
                "competition_person"
            LEFT OUTER JOIN
                "competition_simplescorematchstatistic" ON
                ("competition_person"."uuid" = "competition_simplescorematchstatistic"."player_id")
            LEFT OUTER JOIN
                "competition_match" ON
                ("competition_simplescorematchstatistic"."match_id" = "competition_match"."id")
            LEFT OUTER JOIN
                "competition_teamassociation" ON (
                    "competition_person"."uuid" = "competition_teamassociation"."person_id"
                    AND (
                        "competition_match"."home_team_id" = "competition_teamassociation"."team_id"
                        OR "competition_match"."away_team_id" = "competition_teamassociation"."team_id"
                    )
                )
            LEFT OUTER JOIN
                "%(user)s" ON ("competition_person"."user_id" = "%(user)s"."id")
            WHERE
                "competition_person"."club_id" = %%(club)s
            GROUP BY
                "competition_person"."uuid",
                "competition_person"."first_name",
                "competition_person"."last_name",
                "competition_person"."gender",
                "competition_person"."date_of_birth",
                "competition_person"."club_id",
                "competition_person"."user_id",
                "%(user)s"."id"
            ORDER BY
                "competition_person"."last_name" ASC,
                "competition_person"."first_name" ASC
        """ % {
            "user": get_user_model()._meta.db_table
        }
        params = {
            "club": self.pk,
        }
        members = FauxQueryset(Person)
        for member in Person.objects.raw(query, params):
            members.append(member)
        res = {
            "members": members,
        }
        return res

    @cached_property
    def _mvp_select_related(self):
        res = {
            "teams": [
                "division__season__competition",
            ],
        }
        return res

    @cached_property
    def _mvp_only(self):
        res = {
            "teams": [
                "club",
                "title",
                "division__title",
                "division__short_title",
                "division__season__title",
                "division__season__short_title",
                "division__season__start_date",
                "division__season__competition__title",
                "division__season__competition__short_title",
                "division__season__competition__order",
            ],
        }
        return res

    def _get_admin_namespace(self):
        return "admin:fixja:club"

    def __str__(self):
        return self.title

    @property
    def matches(self):
        home = Match.objects.filter(home_team__club=self).select_related(
            "home_team", "away_team"
        )
        away = Match.objects.filter(away_team__club=self).select_related(
            "home_team", "away_team"
        )
        return home | away


class RankDivision(AdminUrlMixin, OrderedSitemapNode):
    enabled = BooleanField(default=True)

    class Meta(OrderedSitemapNode.Meta):
        pass


class RankTeam(models.Model):
    club = models.ForeignKey(Club, on_delete=CASCADE)
    division = models.ForeignKey(RankDivision, on_delete=CASCADE)

    class Meta:
        ordering = ("division",)
        unique_together = ("club", "division")

    def __str__(self):
        return "%s (%s)" % (self.club, self.division)

    def __repr__(self):
        return "%r, %r" % (self.club, self.division)

    def natural_key(self):
        return "{} {}".format(self.club.title, self.division.title)


class RankPoints(models.Model):
    team = models.ForeignKey(RankTeam, on_delete=CASCADE)
    points = models.DecimalField(default=0, max_digits=6, decimal_places=3)
    date = models.DateField()

    class Meta:
        ordering = ("-date", "team", "-points")

    def __str__(self):
        return "%r, %r, %r" % (self.team, self.points, self.date)


class Person(AdminUrlMixin, models.Model):
    """
    This is not tied to the `django.contrib.auth.models.User` model because
    I don't want to clutter the user-space with hundreds of accounts that may
    never be used for authentication.

    We should however make it easy to clone a `Person` into a `User` at some
    point - not just now though.
    """

    uuid = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)

    club = ForeignKey(
        "competition.Club",
        related_name="members",
        label_from_instance="title",
        on_delete=PROTECT,
    )

    first_name = models.CharField(max_length=50)
    last_name = models.CharField(max_length=50)

    gender = models.CharField(max_length=1, choices=GENDER_CHOICES)
    date_of_birth = DateField(blank=True, null=True)

    email = models.EmailField(blank=True)
    home_phone = models.CharField(max_length=30, blank=True)
    work_phone = models.CharField(max_length=30, blank=True)
    mobile_phone = models.CharField(max_length=30, blank=True)

    user = models.OneToOneField(
        settings.AUTH_USER_MODEL, blank=True, null=True, on_delete=SET_NULL
    )

    def _get_admin_namespace(self):
        return "admin:fixja:club:person"

    def _get_url_args(self):
        return (self.club_id, self.pk)

    @cached_property
    def _mvp_select_related(self):
        res = {
            "statistics": [
                "match__stage__division__season__competition",
                "match__home_team",
                "match__away_team",
            ],
        }
        return res

    class Meta:
        ordering = ("last_name", "first_name")
        verbose_name_plural = "people"

    def __str__(self):
        return self.get_full_name

    @property
    def get_full_name(self):
        return "{}, {}".format(self.last_name, self.first_name)

    @cached_property
    def age(self):
        if self.date_of_birth:
            age = relativedelta(timezone.now().date(), self.date_of_birth)
            return age.years

    @cached_property
    def teams(self):
        return Team.objects.filter(
            pk__in=self.teamassociation_set.values_list("team_id", flat=True)
        )

    @cached_property
    def can_delete(self):
        if self.statistics.exists():
            return False

        now = timezone.now()
        if self.teamassociation_set.filter(team__division__season__start_date__lt=now):
            return False

        return True

    @cached_property
    def stats(self):
        return self.statistics.aggregate(
            played=Sum("played"),
            points=Sum("points"),
        )


class Season(AdminUrlMixin, RankImportanceMixin, OrderedSitemapNode):
    rank_importance_parent_attr = "competition"

    competition = ForeignKey(Competition, related_name="seasons", on_delete=PROTECT)
    hashtag = models.CharField(
        max_length=30,
        blank=True,
        null=True,
        validators=[validate_hashtag],
        verbose_name="Hash Tag",
        help_text=mark_safe(
            _("Your official <em>hash tag</em> " "for social media promotions.")
        ),
    )
    enabled = BooleanField(default=True)
    start_date = DateField(blank=True, null=True)
    mode = models.IntegerField(
        choices=SEASON_MODE_CHOICES,
        default=WEEKLY,
        help_text=_(
            "Used by the draw wizard to help "
            "you set your match dates & times "
            "automatically."
        ),
    )
    statistics = BooleanField(
        default=True,
        help_text=_(
            "Set to No if you do not wish to "
            "keep scoring or most valuable "
            "player statistics."
        ),
    )
    mvp_results_public = DateTimeField(
        blank=True,
        null=True,
        verbose_name=_("MVP public at"),
        help_text=_(
            "The time when the results "
            "of the MVP voting will be "
            "made public on the "
            "website. Leave blank to "
            "show at all times."
        ),
    )

    live_stream = BooleanField(default=False)
    live_stream_privacy = models.CharField(
        max_length=20,
        choices=LiveStreamPrivacy.choices,
        default=LiveStreamPrivacy.PRIVATE,
    )

    complete = BooleanField(
        default=False,
        help_text=_(
            "Set to indicate this season is in "
            "the past. Optimises progression "
            "calculations."
        ),
    )
    disable_calendar = BooleanField(
        default=False,
        help_text=_(
            "Set to prevent the iCalendar feature "
            "for this season. Will hide icon in "
            "front-end and disable functionality. "
            "Batch process may disable after last "
            "match of tournament has taken place."
        ),
    )
    timezone = TimeZoneField(max_length=50, blank=True, null=True)

    forfeit_notifications = ManyToManyField(
        settings.AUTH_USER_MODEL,
        blank=True,
        related_name=None,
        help_text=_(
            "When a team advises they are forfeiting, notify the "
            "opposition team plus these people."
        ),
    )

    class Meta(OrderedSitemapNode.Meta):
        unique_together = (
            ("title", "competition"),
            ("slug", "competition"),
        )

    def _get_admin_namespace(self):
        return "admin:fixja:competition:season"

    def _get_url_args(self):
        return (self.competition_id, self.pk)

    def __repr__(self):
        return "<Season: {} - {}>".format(self.competition, self)

    @property
    def datetimes(self):
        if not hasattr(self, "_dates"):
            queryset = self.matches.exclude(
                Q(date__isnull=True)
                | Q(time__isnull=True)
                | Q(play_at__timezone__isnull=True)
            )
            values_list = queryset.values_list(
                "date", "time", "play_at__timezone"
            ).distinct()
            self._dates = sorted([combine_and_localize(*v) for v in values_list])
        return self._dates

    @property
    def dates(self):
        return [dt.date() for dt in self.datetimes]

    @property
    def matches(self):
        return Match.objects.filter(stage__division__season=self).select_related(
            "home_team", "away_team"
        )

    @property
    def clubs(self):
        ids = (
            Team.objects.filter(division__season=self)
            .values_list("club", flat=True)
            .distinct()
        )
        return self.competition.clubs.filter(pk__in=ids)

    def get_places(self):
        venues = self.venues.values_list("pk", flat=True)
        grounds = Ground.objects.filter(venue__in=venues)
        pks = (
            set(venues)
            .difference(grounds.values_list("venue", flat=True))
            .union(grounds.values_list("pk", flat=True))
        )
        return Place.objects.filter(pk__in=pks)

    def get_timeslots(self, date=None):
        # work out the timeslot rules to exclude
        exc = Q()
        if date is not None:
            # exclude rules that expired on a day earlier than the
            # specified date
            exc |= Q(end_date__isnull=False, end_date__lt=date)
            # exclude rules that start after the date specified
            exc |= Q(start_date__isnull=False, start_date__gt=date)
        # build our ruleset and consume it as a list
        rset = rruleset()
        for timeslot in self.timeslots.exclude(exc):
            rset.rrule(timeslot.rrule())
        return [dt.time() for dt in rset]


class Place(AdminUrlMixin, OrderedSitemapNode):
    abbreviation = models.CharField(max_length=20, blank=True, null=True)
    latlng = LocationField(max_length=100)
    timezone = TimeZoneField(max_length=50, blank=True, null=True)

    class Meta(OrderedSitemapNode.Meta):
        pass

    @property
    def location(self):
        if not hasattr(self, "_latlng_pieces"):
            try:
                self._latlng_pieces = [Decimal(p) for p in self.latlng.split(",")]
            except InvalidOperation:
                self._latlng_pieces = None
        return self._latlng_pieces

    @property
    def latitude(self):
        return self.location and self.location[0]

    @property
    def longitude(self):
        return self.location and self.location[1]

    @property
    def zoom(self):
        return self.location and self.location[2]


class Venue(Place):
    """
    A venue where a season will be played. May have a number of grounds; see
    below. A match will be scheduled to take place on either at a venue or
    on a particular ground.
    """

    season = ForeignKey(Season, related_name="venues", on_delete=PROTECT)

    def _get_admin_namespace(self):
        return "admin:fixja:competition:season:venue"

    def _get_url_args(self):
        return (self.season.competition_id, self.season_id, self.pk)


class Ground(Place):
    """
    A ground (in some sports it may be a court or field) is an individual
    playing surface at a given venue.
    """

    venue = ForeignKey(Venue, related_name="grounds", on_delete=PROTECT)
    live_stream = BooleanField(default=False)
    external_identifier = models.CharField(
        max_length=50, blank=True, null=True, unique=True, db_index=True
    )
    stream_key = models.CharField(
        max_length=50, blank=True, null=True, unique=True, db_index=True
    )

    def _get_admin_namespace(self):
        return "admin:fixja:competition:season:venue:ground"

    def _get_url_args(self):
        return (
            self.venue.season.competition_id,
            self.venue.season_id,
            self.venue_id,
            self.pk,
        )


class Division(
    AdminUrlMixin,
    ModelDiffMixin,
    RankDivisionMixin,
    RankImportanceMixin,
    OrderedSitemapNode,
):
    """
    A model that represents a division within a competition.
    """

    rank_importance_parent_attr = "season"

    season = ForeignKey(Season, related_name="divisions", on_delete=PROTECT)

    points_formula = LadderPointsField(
        blank=True, null=True, verbose_name=_("Points system")
    )
    bonus_points_formula = models.TextField(blank=True, null=True)

    forfeit_for_score = models.SmallIntegerField(null=True)
    forfeit_against_score = models.SmallIntegerField(null=True)
    include_forfeits_in_played = BooleanField(default=True)

    games_per_day = models.SmallIntegerField(
        null=True,
        help_text=_(
            "In Tournament mode, specify how many matches "
            "per day should be scheduled by the automatic "
            "draw generator."
        ),
    )

    draft = BooleanField(
        default=False,
        help_text=_(
            "Marking a division as draft will "
            "prevent matches from being visible "
            "in the front-end."
        ),
    )

    # This is an advanced feature, we would not wish to surface it under
    # normal circumstances, but the theory is that we can use the report URL
    # to construct the minimum data for a division.
    sportingpulse_url = models.URLField(
        max_length=1024, blank=True, null=True, editable=False
    )

    objects = DivisionQuerySet.as_manager()

    class Meta(OrderedSitemapNode.Meta):
        unique_together = (
            ("title", "season"),
            ("slug", "season"),
        )

    def _get_admin_namespace(self):
        return "admin:fixja:competition:season:division"

    def _get_url_args(self):
        return (self.season.competition_id, self.season_id, self.pk)

    @property
    def matches(self):
        return Match.objects.filter(stage__division=self).select_related(
            "home_team", "away_team"
        )

    def ladders(self):
        res = collections.OrderedDict()
        for stage in self.stages.exclude(keep_ladder=False).annotate(
            pool_count=Count("pools")
        ):
            res.update(stage.ladders())
        return res

    def matches_by_date(self):
        res = collections.OrderedDict()
        for stage in self.stages.all():
            res.update(stage.matches_by_date())
        return res


class Stage(AdminUrlMixin, RankImportanceMixin, OrderedSitemapNode):
    rank_importance_parent_attr = "division"

    division = ForeignKey(
        Division, related_name="stages", label_from_instance="title", on_delete=PROTECT
    )
    follows = ForeignKey(
        "self",
        blank=True,
        null=True,
        on_delete=SET_NULL,
        related_name="preceeds",
        label_from_instance="title",
        help_text=mark_safe(
            _(
                "When progressing teams into this stage, which earlier stage "
                "should be used for determining positions.<br>"
                "Default is the immediately preceeding stage."
            )
        ),
    )

    keep_ladder = BooleanField(
        default=True,
        verbose_name="Keep a ladder",
        help_text=mark_safe(
            _(
                "Set this to <b>No</b> if this stage does not need to keep a "
                "competition ladder.<br>Usually set to No for a Final "
                "Series or a Knockout stage."
            )
        ),
    )

    scale_group_points = BooleanField(
        default=False,
        help_text=mark_safe(
            _(
                "In stages with multiple pools, adjust points in "
                "the smaller groups to compensate for the reduced "
                "opportunity to score points.<br>You "
                "<strong>should</strong> also set 0 points for Bye "
                "matches."
            )
        ),
    )

    carry_ladder = BooleanField(
        default=False,
        verbose_name="Carry over points",
        help_text=mark_safe(
            _(
                "Set this to <b>Yes</b> if this stage should carry over values "
                "from the previous stage."
            )
        ),
    )

    keep_mvp = BooleanField(
        default=True,
        verbose_name="Keep MVP stats",
        help_text=mark_safe(
            _(
                "Set this to <b>No</b> if this stage does not need to keep "
                "track of MVP points.<br>Usually set to No for a Final Series."
            )
        ),
    )

    matches_needing_printing = ManyToManyField(
        "Match", blank=True, related_name="to_be_printed"
    )

    objects = StageQuerySet.as_manager()

    class Meta(OrderedSitemapNode.Meta):
        unique_together = (
            ("title", "division"),
            ("slug", "division"),
        )

    def _get_admin_namespace(self):
        return "admin:fixja:competition:season:division:stage"

    def _get_url_args(self):
        return (
            self.division.season.competition_id,
            self.division.season_id,
            self.division_id,
            self.pk,
        )

    def __str__(self):
        return self.title

    @property
    def teams(self):
        if self.order > 1:
            team_ids = set()
            for (home_team, away_team) in self.matches.values_list(
                "home_team", "away_team"
            ):
                if home_team is not None:
                    team_ids.add(home_team)
                if away_team is not None:
                    team_ids.add(away_team)
            return self.division.teams.filter(pk__in=team_ids)
        else:
            return self.division.teams

    @property
    def comes_after(self):
        if self.follows:
            return self.follows
        after = self.division.stages.filter(order__lt=self.order)
        if after:
            return after.latest("order")
        raise Stage.DoesNotExist

    def ladders(self):
        res = collections.OrderedDict()
        if self.pools.count():
            for pool in self.pools.all():
                res.setdefault(self, collections.OrderedDict()).update(pool.ladders())
        else:
            summary = self.ladder_summary.select_related("team__club")
            res.setdefault(self, summary)
        return res

    def matches_by_date(self):
        tzinfo = timezone.get_current_timezone()
        res = collections.OrderedDict()
        for match in (
            self.matches.select_related(
                "play_at",
                "stage__division",
                "home_team__club",
                "home_team__division",
                "away_team__club",
                "away_team__division",
            )
            .annotate(
                statistics_count=Count("statistics"),
                videos_count=Count("videos"),
                referee_count=Count("referees"),
            )
            .order_by("datetime", "date", "time", "round")
        ):
            res.setdefault(self, collections.OrderedDict()).setdefault(
                match.get_date(tzinfo), []
            ).append(match)
        return res


class StageGroup(AdminUrlMixin, RankImportanceMixin, OrderedSitemapNode):
    """
    A model which represents a sub-grouping of a division; often called a
    'pool', but could also be used to represent a combined division.
    """

    rank_importance_parent_attr = "stage"

    stage = ForeignKey(Stage, related_name="pools", on_delete=PROTECT)
    carry_ladder = BooleanField(
        default=False,
        verbose_name="Carry over points",
        help_text=mark_safe(
            _(
                "Set this to <b>Yes</b> if the ladder for this pool should "
                "carry over values from the previous stage.<br>"
                "Will only apply for matches played against teams that are "
                "now in this group."
            )
        ),
    )

    class Meta(OrderedSitemapNode.Meta):
        verbose_name = "pool"
        unique_together = ("stage", "order")

    def _get_admin_namespace(self):
        return "admin:fixja:competition:season:division:stage:stagegroup"

    def _get_url_args(self):
        return (
            self.stage.division.season.competition_id,
            self.stage.division.season_id,
            self.stage.division_id,
            self.stage_id,
            self.pk,
        )

    @property
    def division(self):
        return self.stage.division

    @property
    def ladder(self):
        team_ids = set()
        for m in self.matches:
            if m.home_team is not None:
                team_ids.add(m.home_team_id)
            if m.away_team is not None:
                team_ids.add(m.away_team_id)
        return LadderSummary.objects.filter(stage=self.stage, team__in=team_ids)

    def ladders(self):
        return {self: self.ladder_summary.select_related("team__club")}

    def matches_by_date(self):
        tzinfo = timezone.get_current_timezone()
        res = collections.OrderedDict()
        matches = self.matches.select_related(
            "play_at",
            "stage__division",
            "home_team__club",
            "home_team__division",
            "away_team__club",
            "away_team__division",
        )
        for match in matches.annotate(
            statistics_count=Count("statistics"),
            video_count=Count("videos"),
            referee_count=Count("referees"),
        ):
            res.setdefault(self, collections.OrderedDict()).setdefault(
                match.get_date(tzinfo), []
            ).append(match)
        return res


class Team(AdminUrlMixin, RankDivisionMixin, OrderedSitemapNode):
    """
    A model which represents a team in a competition. A team may not yet be
    placed into a division, as it might only be at the nomination stage.

    A team has a name which will usually be set by the team manager, but may
    over-ruled and locked by an administrator (for example, to remove a name
    which contains profanity) on a per-team basis.
    """

    rank_division_parent_attr = "division"

    names_locked = BooleanField(
        default=False,
        help_text=mark_safe(
            _(
                "When the team name is locked, the team manager will not be "
                "able to change their team name.<br>"
                "As a tournament manager you can always change the names."
            )
        ),
    )

    club = ForeignKey(
        Club,
        blank=True,
        null=True,
        on_delete=SET_NULL,
        related_name="teams",
        label_from_instance="title",
    )
    division = ForeignKey(
        Division,
        blank=True,
        null=True,
        related_name="teams",
        label_from_instance="title",
        on_delete=PROTECT,
    )
    stage_group = ForeignKey(
        StageGroup,
        verbose_name=_("Pool"),
        blank=True,
        null=True,
        on_delete=SET_NULL,
        related_name="teams",
        label_from_instance="title",
    )

    timeslots_after = TimeField(
        blank=True,
        null=True,
        verbose_name=_("Start after"),
        help_text=_(
            "Specify the earliest time that "
            "this team can play. Leave blank "
            "for no preference."
        ),
    )
    timeslots_before = TimeField(
        blank=True,
        null=True,
        verbose_name=_("Start before"),
        help_text=_(
            "Specify the latest time that "
            "this team can play. Leave blank "
            "for no preference."
        ),
    )
    team_clashes = ManyToManyField(
        "self",
        blank=True,
        verbose_name=_("Don't clash"),
        label_from_instance=team_and_division,
        symmetrical=True,
        help_text=_("Select any teams that must " "not play at the same time."),
    )

    class Meta:
        ordering = (
            "-division__season__start_date",
            "division__order",
            "stage_group__order",
            "order",
        )
        unique_together = (("title", "division"),)

    def clean(self):
        errors = {}

        # Ensure the Meta.unique_together constraint is applied consistently
        other_teams = self.division.teams.exclude(pk=self.pk)
        if other_teams.filter(title__iexact=self.title):
            errors.setdefault("title", []).append(
                _("Team name must be unique in this division.")
            )

        if errors:
            raise ValidationError(errors)

        return super(Team, self).clean()

    def _get_admin_namespace(self):
        return "admin:fixja:competition:season:division:team"

    def _get_url_args(self):
        return (
            self.division.season.competition_id,
            self.division.season_id,
            self.division_id,
            self.pk,
        )

    def __cmp__(self, other):
        if self.division_id == other.division_id:
            return cmp(self.order, other.order)
        return cmp(self.division, other.division)

    def __repr__(self):
        return "<Team: %s>" % self.title

    def __str__(self):
        return self.title

    @property
    def matches(self):
        related = (
            "stage__division",
            "stage_group",
            "home_team__club",
            "away_team__club",
            "home_team__division",
            "away_team__division",
        )
        home = self.home_games.select_related(*related)
        away = self.away_games.select_related(*related)
        return home | away

    def future(self, offset=None):
        if offset is None:
            offset = 15  # FIXME add a value on a season
        dt = timezone.now() - timedelta(minutes=offset)
        d, t = dt.date(), dt.time()
        return self.matches.filter((Q(date__exact=d) & Q(time__gte=t)) | Q(date__gt=d))

    def past(self, offset=None):
        if offset is None:
            offset = 15  # FIXME add a value on a season
        dt = timezone.now() - timedelta(minutes=offset)
        d, t = dt.date(), dt.time()
        return self.matches.filter((Q(date__exact=d) & Q(time__lte=t)) | Q(date__lt=d))

    def next_match(self):
        matches = self.future()
        if matches.count():
            match = matches[0]
            match.next = True
            return match

    def last_match(self):
        matches = self.past()
        if matches.count():
            match = matches.reverse()[0]
            match.last = True
            return match

    def ladders(self):
        res = collections.OrderedDict()
        for stage in self.division.stages.filter(
            keep_ladder=True, ladder_summary__team=self
        ).annotate(pool_count=Count("pools")):
            if stage.pool_count:
                for pool in stage.pools.filter(ladder_summary__team=self):
                    summary = pool.ladder_summary.select_related(
                        "team__club", "team__division"
                    )
                    res.setdefault(stage, collections.OrderedDict()).setdefault(
                        pool, summary
                    )
            else:
                summary = stage.ladder_summary.select_related(
                    "team__club", "team__division"
                )
                res.setdefault(stage, summary)
        return res

    def matches_by_date(self):
        tzinfo = timezone.get_current_timezone()
        res = collections.OrderedDict()
        matches = self.matches.select_related(
            "play_at",
            "stage__division",
            "stage_group",
            "home_team__club",
            "home_team__division",
            "away_team__club",
            "away_team__division",
        )
        for m in matches.annotate(
            statistics_count=Count("statistics"), videos_count=Count("videos")
        ):
            res.setdefault(m.get_date(tzinfo), []).append(m)
        return res


class ByeTeam(object):
    def __init__(self, title="Bye", *args, **kwargs):
        super(ByeTeam, self).__init__(*args, **kwargs)
        self.pk = None
        self.title = mark_safe(
            '<span class="bye" title="%s">%s</span>' % (title, title)
        )

    def __nonzero__(self):
        return False

    def __str__(self):
        return self.title

    def save(self, *args, **kwargs):
        raise NotImplementedError


class UndecidedTeam(AdminUrlMixin, models.Model):
    """
    A model which represents a team in a competition that is dependant on the
    results of an earlier `Stage`. This type of entry fills a void in a draw by
    being assigned as a reference in a match. Upon "progressing" teams at the
    conclusion of a stage, these links will be updated by listeners to signals.
    """

    formula = models.CharField(max_length=20, blank=True)
    label = models.CharField(max_length=30, blank=True)
    stage = ForeignKey(
        Stage,
        related_name="undecided_teams",
        label_from_instance="title",
        on_delete=PROTECT,
    )
    stage_group = ForeignKey(
        StageGroup,
        blank=True,
        null=True,
        on_delete=SET_NULL,
        related_name="undecided_teams",
        verbose_name=_("Pool"),
        label_from_instance="title",
    )

    class Meta:
        ordering = ("stage_group", "formula")
        # verbose_name = 'team'

    def __str__(self):
        return self.title

    def _get_admin_namespace(self):
        return "admin:fixja:competition:season:division:stage:undecidedteam"

    def _get_url_args(self):
        return (
            self.stage.division.season.competition_id,
            self.stage.division.season_id,
            self.stage.division_id,
            self.stage_id,
            self.pk,
        )

    @cached_property
    def choices(self):
        if self.formula:
            __, group, __ = stage_group_position(self.stage, self.formula)
            if group is not None:
                return group.teams
        return self.stage.division.teams

    @property
    def title(self):
        if not self.formula and self.label:
            return self.label

        stage, group, position = stage_group_position(self.stage, self.formula)

        c = {
            "stage": stage,
            "group": group,
            "position": position,
        }

        return stage_group_position_tpl.render(c).strip()

    @property
    def matches(self):
        home = self.home_games.select_related("home_team", "away_team")
        away = self.away_games.select_related("home_team", "away_team")
        return home | away


class ClubRole(AdminUrlMixin, models.Model):
    competition = ForeignKey(
        Competition,
        related_name="club_roles",
        label_from_instance="title",
        on_delete=PROTECT,
    )
    name = models.CharField(max_length=50)

    class Meta:
        ordering = ("name",)

    def __str__(self):
        return self.name

    def _get_admin_namespace(self):
        return "admin:fixja:competition:clubrole"

    def _get_url_args(self):
        return (self.competition_id, self.pk)


class ClubAssociation(AdminUrlMixin, models.Model):
    club = ForeignKey(
        Club, related_name="staff", label_from_instance="title", on_delete=PROTECT
    )
    person = ForeignKey(Person, label_from_instance="get_full_name", on_delete=PROTECT)
    roles = ManyToManyField(ClubRole, label_from_instance="name")

    class Meta:
        ordering = ("person__last_name", "person__first_name")
        unique_together = ("club", "person")
        verbose_name = _("Official")
        verbose_name_plural = _("Officials")

    def _get_admin_namespace(self):
        return "admin:fixja:club:clubassociation"

    def _get_url_args(self):
        return (self.club_id, self.pk)


class TeamRole(AdminUrlMixin, models.Model):
    competition = ForeignKey(
        Competition,
        related_name="team_roles",
        label_from_instance="title",
        on_delete=PROTECT,
    )
    name = models.CharField(max_length=50)

    class Meta:
        ordering = ("name",)

    def __str__(self):
        return self.name

    def _get_admin_namespace(self):
        return "admin:fixja:competition:teamrole"

    def _get_url_args(self):
        return (self.competition_id, self.pk)


class TeamAssociation(AdminUrlMixin, models.Model):
    team = ForeignKey(Team, related_name="people", on_delete=PROTECT)
    person = ForeignKey(
        Person, null=True, label_from_instance="get_full_name", on_delete=PROTECT
    )
    roles = ManyToManyField(TeamRole, blank=True, label_from_instance="name")
    number = models.IntegerField(blank=True, null=True)
    is_player = BooleanField(default=True)

    def __str__(self):
        return self.person.get_full_name

    class Meta:
        ordering = (
            "-is_player",
            "number",
            "person__last_name",
            "person__first_name",
        )
        unique_together = ("team", "person")
        verbose_name = "linked person"
        verbose_name_plural = "linked people"

    def _get_admin_namespace(self):
        return "admin:fixja:competition:season:division:team:teamassociation"

    def _get_url_args(self):
        return (
            self.team.division.season.competition_id,
            self.team.division.season_id,
            self.team.division_id,
            self.team_id,
            self.pk,
        )

    def statistics(self):
        stats = SimpleScoreMatchStatistic.objects.filter(
            player=self.person, player__teamassociation__team=self.team
        )
        stats = stats.filter(
            Q(match__home_team=self.team) | Q(match__away_team=self.team)
        )
        return stats.aggregate(
            played=Sum("played"),
            points=Sum("points"),
            mvp=Sum("mvp"),
        )


class SeasonReferee(AdminUrlMixin, models.Model):
    season = models.ForeignKey(Season, related_name="referees", on_delete=PROTECT)
    club = models.ForeignKey(Club, related_name="referees", on_delete=PROTECT)
    person = ForeignKey(Person, label_from_instance="get_full_name", on_delete=PROTECT)

    def __str__(self):
        return self.person.get_full_name

    class Meta:
        ordering = (
            "season",
            "person__last_name",
            "person__first_name",
        )
        unique_together = ("season", "person")
        verbose_name = _("referee")

    def _get_admin_namespace(self):
        return "admin:fixja:competition:season:seasonreferee"

    def _get_url_args(self):
        return (self.season.competition_id, self.season_id, self.pk)


class SeasonAssociation(AdminUrlMixin, models.Model):
    club = models.ForeignKey(Club, related_name="officials", on_delete=PROTECT)
    season = models.ForeignKey(Season, related_name="officials", on_delete=PROTECT)
    person = ForeignKey(Person, label_from_instance="get_full_name", on_delete=PROTECT)
    roles = ManyToManyField(ClubRole, label_from_instance="name")

    class Meta:
        ordering = (
            "club",
            "season",
            "person__last_name",
            "person__first_name",
        )
        unique_together = ("season", "person")


class Match(AdminUrlMixin, RankImportanceMixin, models.Model):
    uuid = models.UUIDField(
        primary_key=False,
        default=uuid.uuid4,
        editable=False,
        unique=True,
        db_index=True,
    )

    rank_importance_parent_attr = ("stage_group", "stage")

    stage = ForeignKey(
        Stage,
        label_from_instance="title",
        null=True,
        related_name="matches",
        on_delete=PROTECT,
    )
    stage_group = ForeignKey(
        StageGroup,
        blank=True,
        null=True,
        verbose_name=_("Pool"),
        label_from_instance="title",
        related_name="matches",
        on_delete=PROTECT,
    )

    label = models.CharField(max_length=100, blank=True, null=True)

    home_team = ForeignKey(
        Team,
        blank=True,
        null=True,
        related_name="home_games",
        label_from_instance="title",
        on_delete=PROTECT,
    )
    away_team = ForeignKey(
        Team,
        blank=True,
        null=True,
        related_name="away_games",
        label_from_instance="title",
        on_delete=PROTECT,
    )

    referees = ManyToManyField(SeasonReferee, blank=True, related_name="matches")

    # these fields are used when the home/away teams are to be determined by
    # some form of calculation - usually by a position within the ladder or
    # group/pool
    home_team_undecided = ForeignKey(
        UndecidedTeam,
        blank=True,
        null=True,
        related_name="home_games",
        label_from_instance="title",
        on_delete=PROTECT,
    )
    home_team_eval = models.CharField(max_length=10, blank=True, null=True)
    home_team_eval_related = ForeignKey(
        "self",
        blank=True,
        null=True,
        related_name="+",
        label_from_instance="pk",
        on_delete=PROTECT,
    )
    away_team_undecided = ForeignKey(
        UndecidedTeam,
        blank=True,
        null=True,
        related_name="away_games",
        label_from_instance="title",
        on_delete=PROTECT,
    )
    away_team_eval = models.CharField(max_length=10, blank=True, null=True)
    away_team_eval_related = ForeignKey(
        "self",
        blank=True,
        null=True,
        related_name="+",
        label_from_instance="pk",
        on_delete=PROTECT,
    )

    evaluated = models.BooleanField(null=True)

    is_washout = BooleanField(default=False)

    date = DateField(blank=True, null=True)
    time = TimeField(blank=True, null=True)
    datetime = DateTimeField(blank=True, null=True)

    play_at = ForeignKey(
        Place, label_from_instance="title", blank=True, null=True, on_delete=SET_NULL
    )

    home_team_score = models.IntegerField(blank=True, null=True)
    away_team_score = models.IntegerField(blank=True, null=True)

    is_bye = BooleanField(default=False)
    bye_processed = BooleanField(default=False)

    is_forfeit = BooleanField(default=False)
    forfeit_winner = ForeignKey(
        Team,
        blank=True,
        null=True,
        related_name="+",
        label_from_instance="title",
        on_delete=PROTECT,
    )

    round = models.IntegerField(blank=True, null=True)

    include_in_ladder = BooleanField(default=True)

    external_identifier = models.CharField(
        max_length=20, blank=True, null=True, unique=True, db_index=True
    )

    videos = PG.ArrayField(
        models.URLField(),
        null=True,
    )
    live_stream = BooleanField(default=False)
    live_stream_bind = models.CharField(
        max_length=50, blank=True, null=True, db_index=True
    )

    objects = MatchManager()

    class Meta:
        get_latest_by = "datetime"
        ordering = (
            "date",
            "stage",
            "round",
            "is_bye",
            "time",
            "play_at__ground__order",
            "pk",
        )
        verbose_name_plural = "matches"

    def _get_admin_namespace(self):
        return "admin:fixja:competition:season:division:stage:match"

    def _get_url_args(self):
        return (
            self.stage.division.season.competition_id,
            self.stage.division.season_id,
            self.stage.division_id,
            self.stage_id,
            self.pk,
        )

    def get_date(self, tzinfo):
        dt = self.get_datetime(tzinfo)
        if dt is not None:
            return dt.date()
        return self.date

    def get_datetime(self, tzinfo=None):
        res = self.datetime
        if res is None:
            exclude = (
                Q(pk=self.pk)
                | Q(date__isnull=True)
                | Q(time__isnull=True)
                | Q(play_at__isnull=True)
            )
            # identify the matches which are in the same round as this match
            siblings = Match.objects.exclude(exclude).filter(
                stage=self.stage, round=self.round
            )
            # if we're part of a stage group, further filter the sibling list
            if self.stage_group:
                siblings = siblings.filter(stage_group=self.stage_group)
            if siblings:
                res = max([s.datetime for s in siblings])
        if res is not None and settings.USE_TZ and tzinfo is not None:
            return timezone.localtime(res, tzinfo)
        return res

    def forfeit(self, team, actor=None):
        """
        Update the Match to be recorded as a forfeit.

        Optionally provide detail of who notified the forfeit.

        :param team: the home_team or away_team, or None for double forfeit
        :param actor: the person who notified us of the forfeit
        :return: success of the operation
        """
        assert team in (self.home_team, self.away_team, None)

        if actor is not None:
            # If it is a double forfeit, ensure the actor is a member of
            # either the home or away team, or an admin.
            if team is None:
                team_association = self.home_team.people.filter(
                    person=actor
                ) | self.away_team.people.filter(person=actor)
            # Ensure the actor a member of the forfeiting team, or an admin.
            else:
                team_association = team.people.filter(person=actor)

            assert team_association.exists() or actor.is_staff

        # Matches that have already been forfeit can not be forfeit again.
        if self.is_forfeit:
            return False

        # Matches that already have a result can not be forfeit. We should not
        # have partial scores, but check for either to be not None.
        if self.home_team_score is not None or self.away_team_score is not None:
            return False

        self.is_forfeit = True

        # Set the match result based on who has forfeit.
        if self.home_team == team:
            self.forfeit_winner = self.away_team
            self.home_team_score = self.stage.division.forfeit_against_score
            self.away_team_score = self.stage.division.forfeit_for_score
        elif self.away_team == team:
            self.forfeit_winner = self.home_team
            self.home_team_score = self.stage.division.forfeit_for_score
            self.away_team_score = self.stage.division.forfeit_against_score
        else:
            self.home_team_score = self.stage.division.forfeit_against_score
            self.away_team_score = self.stage.division.forfeit_against_score

        self.save()

        # Send the match_forfeit signal so we can handle notifications
        match_forfeit.send(sender=self.__class__, match=self, team=team)

        return True

    def clean(self):
        # Remember, this will only fire if the `Match.save` method is invoked;
        # the `update` queryset method will cause this to be bypassed
        # altogether, so any `update` that changes the `date` or `time` fields
        # (such as rescheduling due to wet weather) should take this into
        # account.
        errors = {}

        if self.date is not None:
            if self.stage.division.season.start_date:
                if self.stage.division.season.start_date > self.date:
                    errors.setdefault("date", []).append(
                        _("This date is before the start of the season.")
                    )
            if self.stage.division.season.exclusions.filter(date=self.date):
                errors.setdefault("date", []).append(
                    _("This date has been excluded for this season.")
                )
            if self.stage.division.exclusions.filter(date=self.date):
                errors.setdefault("date", []).append(
                    _("This date has been excluded for this division.")
                )

        if self.stage_group and self.stage_group not in self.stage.pools.all():
            errors.setdefault("stage_group", []).append(
                _("This pool is not in the selected division")
            )

        if errors:
            raise ValidationError(errors)

        try:
            self.datetime = datetime.combine(self.date, self.time)
        except TypeError:
            self.datetime = None

        if self.datetime is not None and settings.USE_TZ:
            try:
                tzinfo = pytz.timezone(self.play_at.timezone)
            except (AttributeError, pytz.UnknownTimeZoneError):
                tzinfo = timezone.get_default_timezone()
            else:
                self.datetime = timezone.make_aware(self.datetime, tzinfo)

    @cached_property
    def title(self):
        context = {
            "home": self.get_home_team(),
            "away": self.get_away_team(),
        }
        return match_title_tpl.render(context)

    def _get_team(self, field, plain=False):
        team = getattr(self, field)
        if team:
            return team

        team_undecided = getattr(self, f"{field}_undecided")
        if team_undecided:
            team_eval = team_undecided.formula
            team_eval_related = None
        else:
            team_eval = getattr(self, f"{field}_eval")
            team_eval_related = getattr(self, f"{field}_eval_related")

        try:
            stage, group, position = stage_group_position_re.match(team_eval).groups()
        except (AttributeError, TypeError):
            if not team_undecided and self.is_bye:
                return ByeTeam()
            stage = group = position = None
        else:
            stage = self.stage.comes_after

        if team_undecided and stage is None:
            if plain:
                return str(team_undecided)
            return {"title": str(team_undecided)}
        elif team_eval in WIN_LOSE:
            context = {
                "position": WIN_LOSE[team_eval],
                "match": team_eval_related,
            }
            template = win_lose_team_tpl
        else:
            context = {
                "position": position,
                "stage": stage,
            }
            template = stage_group_position_tpl
            if group is not None:
                if stage.pools.count():
                    context["group"] = stage.pools.all()[int(group) - 1]
                else:
                    context["group"] = {
                        "title": mark_safe('<span class="error">ERROR</span>'),
                    }
                    context.setdefault("errors", []).append("Invalid group.")
        if plain:
            return template.render(context).strip()
        return {"title": template.render(context).strip()}

    def get_home_team(self):
        return self._get_team("home_team")

    def get_away_team(self):
        return self._get_team("away_team")

    def get_home_team_plain(self):
        return self._get_team("home_team", True)

    def get_away_team_plain(self):
        return self._get_team("away_team", True)

    def _winner_loser(self, win_lose):
        winner = loser = None

        if self.home_team_score is not None and self.away_team_score is not None:
            if self.is_forfeit:
                winner = self.forfeit_winner
            elif self.home_team_score > self.away_team_score:
                winner, loser = self.home_team, self.away_team
            elif self.home_team_score < self.away_team_score:
                winner, loser = self.away_team, self.home_team

        if win_lose == "W":
            return winner
        elif win_lose == "L":
            return loser
        else:
            return None

    winner = property(lambda self: self._winner_loser("W"))
    loser = property(lambda self: self._winner_loser("L"))

    def eval(self, lazy=True):
        """
        Attempt to populate the `home_team` and `away_team` fields as
        appropriate.

        When lazy=False we should always evaluate the teams, if possible.
        """
        try:
            stage = self.stage.comes_after
        except Stage.DoesNotExist:
            logger.warn(
                "Stage is first, %r should not be attempting to be " "evaluated.", self
            )
            return (self.home_team, self.away_team)

        positions = {
            index + 1: team
            for index, team in enumerate(
                stage.ladder_summary.values_list("team", flat=True)
            )
        }
        group_positions = {
            index + 1: [each.team for each in group.ladder]
            for index, group in enumerate(stage.pools.all())
        }
        res = [None, None]
        for index, field in enumerate(("home_team", "away_team")):
            team = self._get_team(field)
            if not lazy or not isinstance(team, Team):
                team_undecided = getattr(self, f"{field}_undecided")
                if team_undecided:
                    team_eval = team_undecided.formula
                    team_eval_related = None
                else:
                    team_eval = getattr(self, f"{field}_eval")
                    team_eval_related = getattr(self, f"{field}_eval_related")
                if team_eval in WIN_LOSE:
                    team = team_eval_related._winner_loser(team_eval)
                else:
                    try:
                        stage, group, position = stage_group_position_re.match(
                            team_eval
                        ).groups()
                    except (AttributeError, TypeError):
                        pass
                    else:
                        try:
                            try:
                                g = int(group)
                                p = int(position)
                                team = group_positions[g][p - 1]
                            except TypeError:
                                team = positions[int(position)]
                        except (IndexError, KeyError):
                            pass
                res[index] = team
        return tuple(res)

    def __str__(self):
        return self.title

    def __repr__(self):
        return "<Match: %s: %s>" % (self.round, self)


class LadderBase(models.Model):
    played = models.SmallIntegerField(default=0)
    win = models.SmallIntegerField(default=0)
    loss = models.SmallIntegerField(default=0)
    draw = models.SmallIntegerField(default=0)
    bye = models.SmallIntegerField(default=0)
    forfeit_for = models.SmallIntegerField(default=0)
    forfeit_against = models.SmallIntegerField(default=0)
    score_for = models.SmallIntegerField(default=0)
    score_against = models.SmallIntegerField(default=0)
    bonus_points = models.SmallIntegerField(default=0)
    points = models.DecimalField(default=0, max_digits=6, decimal_places=3)

    class Meta:
        abstract = True

    def __repr__(self):
        return "<%s: #%s>" % (self.__class__.__name__, self.pk)


class LadderEntry(LadderBase):
    match = ForeignKey(Match, related_name="ladder_entries", on_delete=CASCADE)
    team = ForeignKey(
        Team,
        related_name="ladder_entries",
        label_from_instance="title",
        on_delete=CASCADE,
    )
    opponent = ForeignKey(Team, null=True, on_delete=PROTECT)

    objects = LadderEntryManager()

    def __repr__(self):
        return "<LadderEntry: #%s>" % self.pk


class LadderSummary(LadderBase):
    stage = ForeignKey(Stage, related_name="ladder_summary", on_delete=CASCADE)
    stage_group = ForeignKey(
        StageGroup, null=True, related_name="ladder_summary", on_delete=CASCADE
    )
    team = ForeignKey(Team, related_name="ladder_summary", on_delete=PROTECT)
    difference = models.DecimalField(max_digits=6, decimal_places=3, default=0)
    percentage = models.DecimalField(max_digits=10, decimal_places=2, null=True)

    class Meta:
        ordering = (
            "stage",
            "-points",
            "-difference",
            "-percentage",
            "team__title",
        )
        unique_together = ("stage", "team")


class DrawFormat(AdminUrlMixin, models.Model):
    name = models.CharField(max_length=50)
    text = models.TextField(verbose_name="Formula")
    teams = models.PositiveIntegerField(blank=True, null=True)
    is_final = BooleanField(default=False)

    class Meta:
        ordering = ("teams", "name")

    def _get_admin_namespace(self):
        return "admin:fixja:format"

    def _get_url_args(self):
        return (self.pk,)

    def generator(self, stage, start_date=None):
        from tournamentcontrol.competition.draw import DrawGenerator

        generator = DrawGenerator(stage, start_date)
        generator.parse(self.text)
        return generator

    def __str__(self):
        return self.name


class ExclusionDateBase(AdminUrlMixin, models.Model):
    """
    Abstract base class that stores a date to exclude when generating a draw
    automatically from a DrawGenerator.
    """

    date = DateField()

    class Meta:
        abstract = True
        ordering = ("date",)


class SeasonExclusionDate(ExclusionDateBase):
    season = ForeignKey("Season", related_name="exclusions", on_delete=CASCADE)

    class Meta(ExclusionDateBase.Meta):
        unique_together = ("season", "date")
        verbose_name = "exclusion date"

    def _get_admin_namespace(self):
        return "admin:fixja:competition:season:seasonexclusiondate"

    def _get_url_args(self):
        return (self.season.competition_id, self.season_id, self.pk)


class DivisionExclusionDate(ExclusionDateBase):
    division = ForeignKey("Division", related_name="exclusions", on_delete=CASCADE)

    class Meta:
        unique_together = ("division", "date")
        verbose_name = "exclusion date"

    def _get_admin_namespace(self):
        return "admin:fixja:competition:season:division:divisionexclusiondate"

    def _get_url_args(self):
        return (
            self.division.season.competition_id,
            self.division.season_id,
            self.division_id,
            self.pk,
        )


class MatchTimeBase(models.Model):
    """
    Abstract base class that stores enough information to create a recurring
    rule set.
    """

    start = TimeField()
    interval = models.IntegerField()
    count = models.IntegerField()

    class Meta:
        abstract = True

    def rrule_dtstart(self):
        return timezone.now().replace(
            hour=self.start.hour, minute=self.start.minute, second=0
        )

    def rrule_kwargs(self):
        return {
            "dtstart": self.rrule_dtstart(),  # timestamp to start from
            "interval": self.interval,  # every N minutes
            "count": self.count,  # maximum number of timeslots
        }

    def rrule(self):
        return rrule(MINUTELY, **self.rrule_kwargs())


class SeasonMatchTime(AdminUrlMixin, MatchTimeBase):
    season = ForeignKey("Season", related_name="timeslots", on_delete=CASCADE)
    start_date = DateField(verbose_name="From", blank=True, null=True)
    end_date = DateField(verbose_name="Until", blank=True, null=True)

    class Meta:
        verbose_name = "time slot"

    def _get_admin_namespace(self):
        return "admin:fixja:competition:season:seasonmatchtime"

    def _get_url_args(self):
        return (self.season.competition_id, self.season_id, self.pk)

    def __str__(self):
        return "#{}".format(self.pk)


class MatchScoreSheet(AdminUrlMixin, models.Model):
    uuid = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    match = models.ForeignKey(Match, related_name="scoresheets", on_delete=CASCADE)
    image = CloudinaryField(verbose_name=_("Image"))

    def _get_admin_namespace(self):
        return "admin:fixja:competition:season:division:stage:match:matchscoresheet"

    def _get_url_args(self):
        return (
            self.match.stage.division.season.competition_id,
            self.match.stage.division.season_id,
            self.match.stage.division_id,
            self.match.stage_id,
            self.match_id,
            self.pk,
        )

    def __str__(self):
        return str(self.image)


class MatchStatisticBase(models.Model):
    match = ForeignKey("Match", related_name="statistics", on_delete=PROTECT)
    player = ForeignKey("Person", related_name="statistics", on_delete=PROTECT)
    number = models.IntegerField(blank=True, null=True)

    def team(self):
        if self.match.home_team.people.filter(person=self.player):
            team = self.match.home_team
        elif self.match.away_team.people.filter(person=self.player):
            team = self.match.away_team
        else:
            team = None
        return team

    class Meta:
        abstract = True
        ordering = ("match", "number")
        get_latest_by = "match"


class SimpleScoreMatchStatistic(MatchStatisticBase):
    played = models.SmallIntegerField(
        default=0,
        blank=True,
        validators=[
            validators.MinValueValidator(0),
            validators.MaxValueValidator(1),
        ],
    )
    points = models.SmallIntegerField(_("Points"), blank=True, null=True)
    mvp = models.SmallIntegerField(_("MVP"), blank=True, null=True)

    objects = StatisticQuerySet.as_manager()

    def clean(self):
        errors = {}

        if not self.played and self.points:
            errors["points"] = _("A person who did not play cannot score points.")

        if not self.played and self.mvp:
            errors["mvp"] = _("A person who did not play cannot earn MVP points.")

        if self.played and not self.number:
            errors["number"] = _("A person who played must have a shirt number.")

        if errors:
            raise ValidationError(errors)

    def __repr__(self):
        return "<Stat #%s: %s, %s, number=%s, played=%s>" % (
            self.pk,
            self.match,
            self.player,
            self.number,
            self.played,
        )



from django.urls import include, path
from example_app import sites, views

from touchtechnology.common.sites import AccountsSite
from touchtechnology.news.sites import NewsSite

test_context_processors = sites.TestContextProcessorsSite()
test_date_time_field = sites.TestDateTimeFieldSite()
test_generic_views = sites.TestGenericViewsSite()
test_query_string = sites.TestQueryStringSite()
test_pagination = sites.TestPaginationSite()

accounts = AccountsSite()
news = NewsSite()

urlpatterns = [
    path("context-processors/", test_context_processors.urls),
    path("date-time-field/", test_date_time_field.urls),
    path("generic/", test_generic_views.urls),
    path("query-string/", test_query_string.urls),
    path("pagination/", test_pagination.urls),
    path("accounts/", accounts.urls),  # can't login without it
    path("news/", news.urls),
    path("", include("touchtechnology.common.urls")),
    path("", views.index, name="index"),
]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import touchtechnology.common.db.models
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('competition', '0018_resync_field_and_model_defs'),
    ]

    operations = [
        migrations.AlterField(
            model_name='competition',
            name='clubs',
            field=touchtechnology.common.db.models.ManyToManyField(related_name='competitions', to='competition.Club', blank=True),
        ),
    ]



"""
LiveScore / LiveStream integration API resources.

This is not considered official API and should not be relied upon - it could change
at any time with changing needs.
"""

from dateutil.relativedelta import relativedelta
from rest_framework import serializers, viewsets
from tournamentcontrol.competition import models


class MatchSerializer(serializers.Serializer):

    id = serializers.UUIDField(source="uuid", read_only=True)
    scheduledStartTime = serializers.DateTimeField(source="datetime", read_only=True)
    scheduledEndTime = serializers.SerializerMethodField(read_only=True)
    comp = serializers.ReadOnlyField(source="stage.division.title")
    round = serializers.SerializerMethodField(read_only=True)
    teamA = serializers.SerializerMethodField(read_only=True)
    teamB = serializers.SerializerMethodField(read_only=True)
    location = serializers.CharField(source="play_at", read_only=True)

    def get_round(self, obj):
        return obj.label or str(obj.round)

    def _get_team(self, obj):
        if obj is None:
            return
        if obj.club_id is not None:
            return obj.club.abbreviation
        return obj.title

    def get_teamA(self, obj):
        return self._get_team(obj.home_team)

    def get_teamB(self, obj):
        return self._get_team(obj.away_team)

    def get_scheduledEndTime(self, obj):
        if obj.datetime is not None:
            return obj.datetime + relativedelta(minutes=50)


class MatchViewSet(viewsets.ModelViewSet):
    """
    This endpoint is used by the BIRDI developed application to seed it's
    list of matches for use with the YouTube Event API for live streams.
    """

    lookup_field = "uuid"
    serializer_class = MatchSerializer

    def get_queryset(self):
        return (
            models.Match.objects.select_related(
                "stage__division__season__competition",
                "home_team__club",
                "away_team__club",
                "play_at",
            )
            .filter(
                stage__division__season__slug=self.kwargs["season_slug"],
                stage__division__season__competition__slug=self.kwargs[
                    "competition_slug"
                ],
            )
            .defer(
                # don't select text fields
                "stage__copy",
                "stage__division__copy",
                "stage__division__season__copy",
                "stage__division__season__competition__copy",
                # unused home team fields
                "home_team__club__short_title",
                "home_team__club__slug",
                "home_team__club__email",
                "home_team__club__website",
                "home_team__club__twitter",
                "home_team__club__facebook",
                "home_team__club__youtube",
                "home_team__club__primary",
                "home_team__club__primary_position",
                # unused away team fields
                "away_team__club__short_title",
                "away_team__club__slug",
                "away_team__club__email",
                "away_team__club__website",
                "away_team__club__twitter",
                "away_team__club__facebook",
                "away_team__club__youtube",
                "away_team__club__primary",
                "away_team__club__primary_position",
                # long textual fields that are unused
                "stage__division__points_formula",
                "play_at__latlng",
            )
        )



from django.db import models
from django.db.models import DateTimeField, ManyToManyField
from django.template.defaultfilters import slugify
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from imagekit.models import ImageSpecField

from touchtechnology.admin.mixins import AdminUrlMixin
from touchtechnology.common.db.models import BooleanField, HTMLField
from touchtechnology.news.app_settings import (
    DETAIL_IMAGE_KWARGS,
    DETAIL_IMAGE_PROCESSORS,
    THUMBNAIL_IMAGE_KWARGS,
    THUMBNAIL_IMAGE_PROCESSORS,
)
from touchtechnology.news.image_processors import processor_factory
from touchtechnology.news.query import ArticleQuerySet, CategoryQuerySet


class AdminUrlModel(AdminUrlMixin, models.Model):
    class Meta:
        abstract = True

    def _get_admin_namespace(self):
        return "admin:news:%s" % self._meta.model_name

    def _get_url_args(self):
        return (self.pk,)


class Article(AdminUrlModel):

    headline = models.CharField(max_length=150, verbose_name=_("Headline"))
    slug = models.SlugField(max_length=150, verbose_name=_("Slug"))
    slug_locked = BooleanField(default=False, verbose_name=_("Locked Slug"))

    image = models.ImageField(
        blank=True, null=True, upload_to="news", verbose_name=_("Image")
    )

    abstract = models.TextField(verbose_name=_("Abstract"))

    published = DateTimeField(
        verbose_name=_("Published"),
        help_text=_("Set a date & time in the future to schedule " "an announcement."),
    )

    byline = models.CharField(max_length=75, blank=True, verbose_name=_("Byline"))
    keywords = models.CharField(max_length=255, blank=True, verbose_name=_("Keywords"))
    categories = ManyToManyField("Category", blank=True, verbose_name=_("Categories"))

    is_active = BooleanField(default=True, verbose_name=_("Enabled"))

    copy = HTMLField(_("Copy"), blank=True)

    # image and imagekit spec files
    thumbnail = ImageSpecField(
        source="image",
        processors=processor_factory(THUMBNAIL_IMAGE_PROCESSORS),
        **THUMBNAIL_IMAGE_KWARGS
    )
    detail_image = ImageSpecField(
        source="image",
        processors=processor_factory(DETAIL_IMAGE_PROCESSORS),
        **DETAIL_IMAGE_KWARGS
    )

    last_modified = DateTimeField(auto_now=True)

    objects = ArticleQuerySet.as_manager()

    class Meta:
        ordering = ("-published",)

    def clean(self):
        if not self.slug_locked or not self.slug:
            self.slug = slugify(self.headline)

    def get_absolute_url(self):
        return reverse(
            "news:article",
            kwargs={
                "year": self.published.year,
                "month": self.published.strftime("%b").lower(),
                "day": self.published.day,
                "slug": self.slug,
            },
        )

    def __str__(self):
        return self.headline


class Translation(AdminUrlModel):

    article = models.ForeignKey(
        Article, related_name="translations", on_delete=models.CASCADE
    )
    locale = models.CharField(max_length=10, blank=False, null=False)
    headline = models.CharField(max_length=150, verbose_name=_("Headline"))
    abstract = models.TextField(verbose_name=_("Abstract"))
    copy = HTMLField(_("Copy"), blank=True)

    def __getattr__(self, name):
        try:
            return super().__getattr__(name)
        except AttributeError:
            return getattr(self.article, name)

    def _get_admin_namespace(self):
        return "admin:news:article:%s" % self._meta.model_name

    def _get_url_args(self):
        return (self.article.pk, self.locale)

    def get_absolute_url(self):
        published = self.article.published
        return reverse(
            "news:translation",
            kwargs={
                "year": published.year,
                "month": published.strftime("%b").lower(),
                "day": published.day,
                "slug": self.article.slug,
                "locale": self.locale,
            },
        )

    def __str__(self):
        return self.headline

    class Meta:
        unique_together = ("article", "locale")


class Category(AdminUrlModel):

    title = models.CharField(max_length=75, verbose_name=_("Title"))
    short_title = models.CharField(
        max_length=50,
        blank=True,
        verbose_name=_("Short Title"),
        help_text=_("Used in navigation, a shorter alternative " "to the main title."),
    )
    slug = models.SlugField(max_length=75, verbose_name=_("Slug"))
    slug_locked = BooleanField(default=False, verbose_name=_("Locked Slug"))

    is_active = BooleanField(default=True, verbose_name="Enabled")

    hidden_from_navigation = BooleanField(
        default=False,
        verbose_name=_("Hide from menus"),
        help_text=_(
            "When set to 'Yes' this object will still be available, "
            "but will not appear in menus."
        ),
    )

    last_modified = DateTimeField(auto_now=True)

    objects = CategoryQuerySet.as_manager()

    class Meta:
        verbose_name_plural = "categories"

    def clean(self):
        if not self.slug_locked or not self.slug:
            self.slug = slugify(self.title)

    def get_absolute_url(self):
        return reverse("news:category", kwargs={"category": self.slug})

    def __str__(self):
        return self.title



# Generated by Django 3.2.13 on 2022-07-30 10:23

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("competition", "0041_ground_live_stream"),
    ]

    operations = [
        migrations.AddField(
            model_name="ground",
            name="external_identifier",
            field=models.CharField(
                blank=True, db_index=True, max_length=20, null=True, unique=True
            ),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import touchtechnology.common.db.models
from django.db import migrations
from django.utils import timezone


class Migration(migrations.Migration):

    dependencies = [
        ('common', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='sitemapnode',
            name='last_modified',
            field=touchtechnology.common.db.models.DateTimeField(default=timezone.now(), auto_now=True),
            preserve_default=False,
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('common', '0002_sitemapnode_last_modified'),
    ]

    operations = [
        migrations.DeleteModel(
            name='PythonPackage',
        ),
        migrations.AlterModelOptions(
            name='sitemapnode',
            options={'verbose_name': 'folder', 'ordering': ('tree_id', 'lft')},
        ),
    ]



from django.db.models.query import QuerySet
from django.utils import timezone


class ArticleQuerySet(QuerySet):

    def active(self):
        return self.filter(is_active=True)

    def live(self):
        return self.filter(published__lte=timezone.now()).active()


class CategoryQuerySet(QuerySet):

    def live(self):
        return self.filter(is_active=True)



import os
import socket

from django.utils import timezone
from touchtechnology.common.forms.tz import SelectTimezoneForm


def env(request):
    return {'SITE_ENV': os.environ.get('SITE_ENV', 'dev')}


def query_string(request):
    return {'QUERY_STRING': request.META.get('QUERY_STRING')}


def site(request):
    return {'hostname': socket.gethostname()}


def tz(request):
    return {'select_timezone_form': SelectTimezoneForm(),
            'TIME_ZONE': timezone.get_current_timezone_name()}



# Generated by Django 3.2.13 on 2022-07-30 13:43

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("competition", "0045_ground_stream_key"),
    ]

    operations = [
        migrations.AlterField(
            model_name="match",
            name="live_stream_bind",
            field=models.CharField(blank=True, db_index=True, max_length=50, null=True),
        ),
    ]



import logging
from collections import OrderedDict
from itertools import zip_longest

from django import template
from django.db import models
from django.template import loader
from django.urls import resolve, reverse_lazy
from guardian.shortcuts import get_objects_for_user

from touchtechnology.common.utils import (
    get_all_perms_for_model_cached,
    model_and_manager,
)

logger = logging.getLogger(__name__)
register = template.Library()


@register.inclusion_tag("mvp/list.html", takes_context=True)
def mvp_list(context, queryset, scope=None, template=None):
    model, manager = model_and_manager(queryset)

    request = context.get("request")
    component = context.get("component")

    if scope is None:
        scope = model._meta.model_name

    # Use the template_path resolution method of the AdminComponent to find
    # the right list for inclusion. Use "list.inc.html" by default because
    # "list.html" is reserved for use by generic_list view.
    if template is None:
        paths = component.template_path("list.inc.html", model._meta.model_name)
        # Ensure there is a sensible default so we always find something
        paths.append("mvp/list.inc.html")
        logger.debug("template_search=%r", paths)
        template = loader.select_template(paths)
        template_name = template.template.name
        logger.debug("template_search_result=%r", template_name)

    match = resolve(request.path)
    namespace = match.namespace
    if match.namespace.rsplit(":", 1)[1] != scope:
        namespace = "%s:%s" % (match.namespace, scope)

    perms = get_all_perms_for_model_cached(model)
    global_perms = any([request.user.has_perm(p) for p in perms])

    # If the user does not have any global permissions then adjust the
    # queryset to return only the objects they can act on.
    if not global_perms:
        queryset = get_objects_for_user(
            request.user,
            perms,
            queryset.select_related(),
            use_groups=True,
            any_perm=True,
        )

    # Tidy up the kwargs found using resolve
    kw = match.kwargs.copy()
    kw.pop("admin", None)
    kw.pop("component", None)

    context.update(
        {
            "model": model,
            "template": template,
            "queryset": queryset,
            "object_list": queryset,
            "create": reverse_lazy("%s:add" % namespace, kwargs=kw),
            # Pass to template the name permissions so we can re-use template
            # code to generically list and add/change/delete objects
            "add_perm": "%s.add_%s" % (model._meta.app_label, model._meta.model_name),
            "view_perm": "%s.view_%s" % (model._meta.app_label, model._meta.model_name),
            "change_perm": "%s.change_%s"
            % (model._meta.app_label, model._meta.model_name),
            "delete_perm": "%s.delete_%s"
            % (model._meta.app_label, model._meta.model_name),
        }
    )
    return context


@register.filter
def related(instance, whitelist=None):
    rel = OrderedDict()

    if whitelist is not None:
        rel = OrderedDict(zip_longest([w for w in whitelist if w], []))

    related_objects = [
        f
        for f in instance._meta.get_fields()
        if (f.one_to_many or f.one_to_one) and f.auto_created
    ]

    for ro in related_objects:
        skip_log = 'filter="related", field="%s", reason="%s"'
        try:
            name = ro.get_accessor_name()
        except AttributeError as e:
            logger.debug(skip_log, name, e)
            continue

        if whitelist is not None and name not in rel:
            logger.debug(skip_log, name, "not in whitelist")
            continue

        if isinstance(ro.field, models.ManyToManyField):
            logger.debug(skip_log, name, "m2m")
            continue

        rel[name] = getattr(instance, name)

    # improve performance of related queries
    related = getattr(instance, "_mvp_related", {})
    annotate = getattr(instance, "_mvp_annotate", {})
    select_related = getattr(instance, "_mvp_select_related", {})
    only = getattr(instance, "_mvp_only", {})

    for name, manager in rel.items():
        # Prioritise fully customised queries. If you take this route, you're
        # on your own - make sure all the fields you will use are returned.
        if name in related:
            manager = related[name]
        # Alternatively perform simple select_related and/or annotations.
        else:
            if name in select_related:
                manager = manager.select_related(*select_related[name])
            if name in annotate:
                manager = manager.annotate(**annotate[name])
            if name in only:
                manager = manager.only(*only[name])
        yield (manager, name)


@register.filter
def mvp_list_template(manager, table=None):
    """
    When looping through a list of "related" managers we may need to use an
    alternative template file.

    We'll need to pass in a mapping of manager's to template names.

    :param manager: a model manager
    :param table: a dictionary of model class to template name
    :returns: template path or None
    """
    if table is None:
        table = {}
    return table.get(manager.model)



from typing import Optional

from django import forms
from django.contrib.auth import get_user_model
from django.contrib.auth.forms import AuthenticationForm
from django.contrib.auth.models import Group
from django.contrib.contenttypes.models import ContentType
from django.db.models import QuerySet
from django.forms import BaseFormSet, formset_factory
from django.utils.functional import cached_property
from django.utils.http import urlencode
from django.utils.translation import gettext_lazy as _
from guardian.models import GroupObjectPermission, UserObjectPermission
from guardian.shortcuts import assign_perm, remove_perm
from modelforms.forms import ModelForm

from touchtechnology.common.forms.fields import (
    EmailField,
    ModelMultipleChoiceField,
)
from touchtechnology.common.forms.mixins import (
    BootstrapFormControlMixin,
    PermissionFormSetMixin,
)


class EmailAuthenticationForm(BootstrapFormControlMixin, AuthenticationForm):
    username = EmailField(label=_("Email Address"))

    def __init__(self, *args, **kwargs):
        super(EmailAuthenticationForm, self).__init__(*args, **kwargs)
        for field_name in self.fields:
            self.fields[field_name].widget.attrs["class"] += " text-center"


class RegistrationForm(forms.Form):
    first_name = forms.CharField(max_length=30)
    last_name = forms.CharField(max_length=30)
    email = EmailField()
    password1 = forms.CharField(widget=forms.PasswordInput, label=_("Password"))
    password2 = forms.CharField(widget=forms.PasswordInput, label=_("Password (again)"))

    def clean_email(self):
        UserModel = get_user_model()
        email = self.cleaned_data.get("email")
        if UserModel.objects.filter(email__iexact=email):
            raise forms.ValidationError(
                _(
                    "This email address is already in "
                    "use. Please supply a different "
                    "email address."
                )
            )
        return email

    def clean_password2(self):
        password1 = self.cleaned_data.get("password1")
        password2 = self.cleaned_data.get("password2")
        if password1 and password2:
            if password1 != password2:
                raise forms.ValidationError(
                    _("The two password fields " "didn't match.")
                )
        return password2


class PermissionForm(ModelForm):
    """
    Generic form to be used to assign row-level permissions to an object.
    """

    staff_only: bool = True
    max_checkboxes: int = 7  # Set based on number of rows visible in MVP layout

    def __init__(self, permission, *args, **kwargs):
        super(PermissionForm, self).__init__(*args, **kwargs)
        self.content_type = ContentType.objects.get_for_model(self.instance)
        self.permission = permission

        # determine if we should filter to only show staff
        user_queryset = get_user_model().objects.all()
        if self.staff_only:
            user_queryset = user_queryset.filter(is_staff=True)

        group_queryset = Group.objects.all()

        self.fields["users"] = ModelMultipleChoiceField(
            queryset=user_queryset,
            required=False,
            initial=self._users_with_perms,
            label_from_instance=lambda o: o.get_full_name(),
        )
        self.fields["users"].widget.attrs.setdefault("class", "form-control")

        self.fields["groups"] = ModelMultipleChoiceField(
            queryset=group_queryset,
            required=False,
            initial=self._groups_with_perms,
        )
        self.fields["groups"].widget.attrs.setdefault("class", "form-control")

    def __repr__(self):
        return "<%s: %s?%s>" % (
            self.__class__.__name__,
            self.permission.codename,
            urlencode(
                {
                    "staff_only": self.staff_only,
                    "max_checkboxes": self.max_checkboxes,
                }
            ),
        )

    @cached_property
    def _users_with_perms(self):
        pks = UserObjectPermission.objects.filter(
            permission=self.permission,
            object_pk=self.instance.pk,
            content_type=self.content_type,
        ).values_list("user", flat=True)
        return get_user_model().objects.filter(pk__in=pks)

    @cached_property
    def _groups_with_perms(self) -> QuerySet:
        pks = GroupObjectPermission.objects.filter(
            permission=self.permission,
            object_pk=self.instance.pk,
            content_type=self.content_type,
        ).values_list("group", flat=True)
        return Group.objects.filter(pk__in=pks)

    def save(self, *args, **kwargs):
        if self.has_changed():
            # work out our before and after state
            users_before = self._users_with_perms
            users_after = self.cleaned_data["users"]
            groups_before = self._groups_with_perms
            groups_after = self.cleaned_data["groups"]

            # determine who has had their per object permissions rescinded
            rescind_users = set(users_before).difference(users_after)
            rescind_groups = set(groups_before).difference(groups_after)

            # determine who has had their per object permissions rescinded
            grant_users = set(users_after).difference(users_before)
            grant_groups = set(groups_after).difference(groups_before)

            # rescind and grant access
            for rescind in rescind_users.union(rescind_groups):
                remove_perm(self.permission.codename, rescind, self.instance)

            for grant in grant_users.union(grant_groups):
                assign_perm(self.permission.codename, grant, self.instance)

        return super(PermissionForm, self).save(*args, **kwargs)


def permissionformset_factory(
    model, staff_only: Optional[bool] = None, max_checkboxes: Optional[int] = None
) -> BaseFormSet:
    """
    For the specified Model, return a FormSet class which will allow you to
    control which specific users & groups should have the ability to change
    and delete a model instance.

    The resulting FormSet class will need to be instantiated with an instance
    of the appropriate type.

    The default behaviour is to grant permissions to staff users only.
    Specifying staff_only = False will result in a form that lists all users.

    When there is a large number of users a widget other than checkboxes might
    be preferred to allow for type-ahead searches. We use the Select2 jQuery
    plugin on to of a multiple select widget. Specify the integer number of
    users that will trigger the switch between checkboxes and Select2.

    :param model: Model
    :param staff_only: bool
    :param max_checkboxes: int
    :return: FormSet class
    """

    if staff_only is None:
        staff_only = PermissionForm.staff_only

    if max_checkboxes is None:
        max_checkboxes = PermissionForm.max_checkboxes

    meta = type("Meta", (), {"model": model, "fields": ("id",)})
    form_class = type(
        "%sPermissionForm" % model.__name__,
        (PermissionForm,),
        {
            "Meta": meta,
            "staff_only": staff_only,
            "max_checkboxes": max_checkboxes,
        },
    )
    formset_base = formset_factory(form_class)
    formset_class = type(
        "%sPermissionFormSet" % model.__name__,
        (PermissionFormSetMixin, formset_base),
        {},
    )
    return formset_class



from __future__ import print_function

import argparse
import sys
from datetime import date

from dateutil.parser import parse as date_parse
from django.core.management import BaseCommand
from django.db import transaction
from django.utils.module_loading import import_string


def decay_function(path):
    try:
        return import_string(path)
    except ImportError:
        msg = 'Invalid decay function "{}" specified'.format(path)
        raise argparse.ArgumentTypeError(msg)


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument(
            '--decay',
            default='tournamentcontrol.competition.rank.no_decay',
            type=decay_function,
        )
        parser.add_argument('--start', type=date_parse)
        parser.add_argument('--at', type=date_parse)
        parser.add_argument('--json', type=argparse.FileType('wb'))
        parser.add_argument('--rebuild', action='store_true', default=False)

    def handle(self, start, at, decay, json, rebuild, *args, **kwargs):
        from tournamentcontrol.competition.models import RankPoints
        from tournamentcontrol.competition.rank import rank, json_rank

        if rebuild and at:
            dates = RankPoints.objects.filter(date__gte=at).dates('date', 'day')
        elif rebuild:
            dates = RankPoints.objects.dates('date', 'day')
        elif at is not None:
            dates = [at]
        else:
            dates = [date.today()]

        with transaction.atomic():
            for d in dates:
                if sys.stdout.isatty():
                    print(
                        "Producing ranking points for %s..." % (d.strftime("%Y-%m-%d"),)
                    )
                if not json:
                    rank(start=start, at=d, decay=decay)
                else:
                    json_rank(stream=json, start=start, at=d, decay=decay)



from __future__ import unicode_literals

from urllib.parse import urlparse, urlunparse

from django.template import Context, Template
from django.test.utils import override_settings
from django.utils.http import urlencode
from test_plus import TestCase
from touchtechnology.common.models import SitemapNode

CSSIFY_TEMPLATE = Template("{% load common %}{{ value|cssify }}")
TWITTIFY_TEMPLATE = Template("{% load common %}{{ value|twittify }}")


class CssifyTest(TestCase):
    def test_cssify_str(self):
        context = Context({"value": "some-normal-slug"})
        value = CSSIFY_TEMPLATE.render(context)
        self.assertEqual("some_normal_slug", value)

    def test_cssify_unicode(self):
        context = Context({"value": "some-normal-slug"})
        value = CSSIFY_TEMPLATE.render(context)
        self.assertEqual("some_normal_slug", value)

    def test_cssify_none(self):
        context = Context({"value": None})
        value = CSSIFY_TEMPLATE.render(context)
        self.assertEqual("", value)


class NavigationTest(TestCase):
    maxDiff = None

    @classmethod
    def setUpClass(cls):
        cls.home = SitemapNode.objects.create(title="Home Page", slug="home")
        cls.about = SitemapNode.objects.create(title="About Us", slug="about")
        cls.about_people = SitemapNode.objects.create(
            title="Our People", slug="people", parent=cls.about
        )
        cls.about_work = SitemapNode.objects.create(
            title="Our Work", slug="work", parent=cls.about
        )
        cls.contact = SitemapNode.objects.create(title="Contact Us", slug="contact")
        cls.about_people_gary = SitemapNode.objects.create(
            title="Gary Reynolds", slug="goodtune", parent=cls.about_people
        )
        cls.about_people_fred = SitemapNode.objects.create(
            title="Fred Nurks",
            slug="freddy",
            parent=cls.about_people,
            hidden_from_navigation=True,
        )

    @classmethod
    def tearDownClass(cls):
        pass

    def test_basic(self):
        template = Template("{% load common %}{% navigation %}")
        value = template.render(Context())
        self.assertHTMLEqual(
            value,
            """
            <ul class="navigation">
                <li class="root first NoneType" id="node{}">
                    <a href="/">Home Page</a>
                </li>
                <li class=" has_children NoneType" id="node{}">
                    <a href="/about/">About Us</a>
                </li>
                <li class=" last NoneType" id="node{}">
                    <a href="/contact/">Contact Us</a>
                </li>
            </ul>
            """.format(
                self.home.pk, self.about.pk, self.contact.pk
            ),
        )

    def test_current_node(self):
        template = Template("{% load common %}{% navigation current_node=node %}")
        context = Context({"node": self.contact})
        value = template.render(context)
        self.assertHTMLEqual(
            value,
            """
            <ul class="navigation">
                <li class="root first NoneType" id="node{}">
                    <a href="/">Home Page</a>
                </li>
                <li class=" has_children NoneType" id="node{}">
                    <a href="/about/">About Us</a>
                </li>
                <li class=" current last NoneType" id="node{}">
                    <a href="/contact/">Contact Us</a>
                </li>
            </ul>
            """.format(
                self.home.pk, self.about.pk, self.contact.pk
            ),
        )

    def test_start_stop(self):
        template = Template("{% load common %}{% navigation start_at=0 stop_at=1 %}")
        value = template.render(Context())
        self.assertHTMLEqual(
            value,
            """
            <ul class="navigation">
                <li class="root first NoneType" id="node{}">
                    <a href="/">Home Page</a>
                </li>
                <li class=" has_children NoneType" id="node{}">
                    <a href="/about/">About Us</a>
                    <ul class="navigation">
                        <li class=" has_children first NoneType" id="node{}">
                            <a href="/about/people/">Our People</a>
                        </li>
                        <li class=" last NoneType" id="node{}">
                            <a href="/about/work/">Our Work</a>
                        </li>
                    </ul>
                </li>
                <li class=" NoneType last" id="node{}">
                    <a href="/contact/">Contact Us</a>
                </li>
            </ul>
            """.format(
                self.home.pk,
                self.about.pk,
                self.about_people.pk,
                self.about_work.pk,
                self.contact.pk,
            ),
        )


class TwittifyTest(TestCase):
    def test_twittify(self):
        context = Context({"value": "@goodtune"})
        value = TWITTIFY_TEMPLATE.render(context)
        expected = """
        @<a class="twitter user" target="_blank"
            href="http://twitter.com/goodtune">goodtune</a>
        """
        self.assertHTMLEqual(expected, value)

    def test_twittify_invalid(self):
        context = Context({"value": "goodtune"})
        value = TWITTIFY_TEMPLATE.render(context)
        expected = "goodtune"
        self.assertHTMLEqual(expected, value)


@override_settings(ROOT_URLCONF="example_app.urls")
class QueryStringTest(TestCase):

    fixtures = ["query_string"]

    def setUp(self):
        self.query = dict(year=2013)
        self.url = urlparse(self.reverse("querystring:index"))

    def test_no_filter_no_page(self):
        self.get("querystring:index")

        self.assertResponseContains('<a href="?page=3">3</a>')
        self.assertResponseNotContains('<a href="?page=4">4</a>')

        self.assertResponseContains('<span id="5">5</span>')
        self.assertResponseNotContains('<span id="9">9</span>')

    def test_no_filter_page(self):
        query = dict(page=2)
        url = self.url._replace(query=urlencode(query, True))
        self.get(urlunparse(url))

        self.assertResponseContains('<a href="?page=3">3</a>')
        self.assertResponseNotContains('<a href="?page=4">4</a>')

        self.assertResponseNotContains('<span id="5">5</span>')
        self.assertResponseContains('<span id="9">9</span>')

    def test_filter_no_page(self):
        url = self.url._replace(query=urlencode(self.query, True))
        self.get(urlunparse(url))

        link = '<a href="?year={year}&amp;page={page}">{page}</a>'
        self.assertResponseContains(link.format(page=2, **self.query))
        self.assertResponseNotContains(link.format(page=3, **self.query))

        self.assertResponseContains('<span id="5">5</span>')
        self.assertResponseNotContains('<span id="9">9</span>')

    def test_filter_page(self):
        query = dict(page=2, **self.query)
        url = self.url._replace(query=urlencode(query, True))
        self.get(urlunparse(url))

        link = '<a href="?year={year}&amp;page={page}">{page}</a>'
        self.assertResponseContains(link.format(page=2, **self.query))
        self.assertResponseNotContains(link.format(page=3, **self.query))

        self.assertResponseNotContains('<span id="5">5</span>')
        self.assertResponseContains('<span id="9">9</span>')


@override_settings(ROOT_URLCONF="example_app.urls")
class ContextTest(TestCase):
    def test_env(self):
        self.get("context:env")
        self.assertResponseContains("dev", html=False)

    def test_tz(self):
        self.get("context:tz")
        self.assertResponseContains("UTC", html=False)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import uuid

from django.db import migrations


def initial_uuid(apps, schema_editor):
    Match = apps.get_model('competition', 'Match')

    for m in Match.objects.select_related('stage__division'):
        m.uuid = uuid.UUID(
            '{00000000-%04x-%04x-%04x-000000000000}' % (
                m.stage.division.pk, m.stage.pk, m.pk))
        m.save()


class Migration(migrations.Migration):

    dependencies = [
        ('competition', '0023_competition__add__match_uuid'),
    ]

    operations = [
        migrations.RunPython(
            initial_uuid, reverse_code=migrations.RunPython.noop),

    ]



import logging
from email.utils import formataddr

from django.conf import settings
from django.core.mail import send_mail
from django.template import Context, Template
from tournamentcontrol.competition.calc import BonusPointCalculator, Calculator
from tournamentcontrol.competition.signals.decorators import disable_for_loaddata
from tournamentcontrol.competition.utils import forfeit_notification_recipients

logger = logging.getLogger(__name__)


@disable_for_loaddata
def match_saved_handler(sender, instance, created, *args, **kwargs):
    """
    Function to be called following a Match being saved.

    This should create the necessary LadderEntry objects for the match in
    question (by deleting previous entries) and inserting new values.
    """
    for ladder_entry in instance.ladder_entries.all():
        ladder_entry.delete()

    if instance.is_bye and instance.bye_processed:
        logger.debug('BYE: Match #%s', instance.pk)
        return create_match_ladder_entries(instance)

    elif instance.is_forfeit:
        logger.debug('FORFEIT: Match #%s', instance.pk)
        return create_match_ladder_entries(instance)

    elif instance.home_team_score is not None and instance.away_team_score is not None:
        logger.debug('RESULT: Match #%s', instance.pk)
        return create_match_ladder_entries(instance)

    logger.debug('SKIPPED: Match #%s', instance.pk)


def match_deleted_handler(sender, instance, *args, **kwargs):
    """
    Function to be called prior to a Match being saved.

    This should remove the related LadderEntry objects for the match in
    question.
    """
    instance.ladder_entries.all().delete()


def create_match_ladder_entries(instance):
    home_ladder = create_team_ladder_entry(instance, 'home')
    away_ladder = create_team_ladder_entry(instance, 'away')
    return dict(home_ladder=home_ladder, away_ladder=away_ladder)


def create_team_ladder_entry(instance, home_or_away):
    from tournamentcontrol.competition.models import LadderEntry

    if home_or_away == 'home':
        opponent = 'away'
    elif home_or_away == 'away':
        opponent = 'home'

    team = getattr(instance, home_or_away + '_team', None)
    team_score = getattr(instance, home_or_away + '_team_score') or 0
    opponent_obj = getattr(instance, opponent + '_team', None)
    opponent_score = getattr(instance, opponent + '_team_score') or 0

    ladder_kwargs = dict(match_id=instance.pk,
                         played=1,
                         score_for=team_score,
                         score_against=opponent_score)

    if team is not None:
        ladder_kwargs['team_id'] = team.pk

    if opponent_obj is not None:
        ladder_kwargs['opponent_id'] = opponent_obj.pk

    if instance.is_bye:
        ladder_kwargs['bye'] = 1

    elif instance.is_forfeit:
        if not instance.stage.division.include_forfeits_in_played:
            ladder_kwargs['played'] = 0
        ladder_kwargs['forfeit_for'] = int(team == instance.forfeit_winner)
        ladder_kwargs['forfeit_against'] = int(team != instance.forfeit_winner)

    else:
        ladder_kwargs['win'] = int(team_score > opponent_score)
        ladder_kwargs['loss'] = int(team_score < opponent_score)
        ladder_kwargs['draw'] = int(team_score == opponent_score)

    if team is not None:
        ladder = LadderEntry(**ladder_kwargs)

        calculator = Calculator(ladder)
        calculator.parse(instance.stage.division.points_formula)
        ladder.points = calculator.evaluate()

        if instance.stage.division.bonus_points_formula:
            bonus_calculator = BonusPointCalculator(ladder)
            bonus_calculator.parse(instance.stage.division.bonus_points_formula)
            ladder.bonus_points = bonus_calculator.evaluate()

        ladder.points += ladder.bonus_points
        ladder.save()

        return ladder


def notify_match_forfeit_email(sender, match, team, *args, **kwargs):
    """
    When a match is notified as having been forfeit, send a notification email
    to players in the opposition team and the designated season administrators.

    :param match: the match that was forfeit
    :param team: the team that forfeit the match
    :return: None
    """
    participants, administrators = forfeit_notification_recipients(match)

    context = Context({
        'match': match,
        'team': team,
    })

    # construct and send an email to the players in the opposition
    subject = Template("Your {{ match.time }} game against {{ team.title }} "
                       "has been forfeit").render(context)
    message = ""
    from_email = settings.DEFAULT_FROM_EMAIL
    recipient_list = [formataddr((p.get_full_name(), p.email))
                      for p in participants]

    send_mail(subject, message, from_email, recipient_list)

    # construct and send an email to the competition administrators
    subject = Template("Forfeit: {{ match }} [{{ match.time }}, "
                       "{{ match.date }}, {{ match.play_at }}]").render(context)
    recipient_list = [formataddr((p.get_full_name(), p.email))
                      for p in administrators]

    send_mail(subject, message, from_email, recipient_list)
